#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

class GetTrackCount:
    from vkmusix.aio import async_

    @async_
    async def getTrackCount(self, ownerId: int = None) -> int:
        """
        Подписывается на обновления музыки пользователя или группы.

        `Пример использования`:

        result = client.followOwner(
            ownerId=-28905875,
        )

        print(result)

        :param ownerId: идентификатор owner'а (пользователь или группа). По умолчанию залогиненный пользователь. (``int``, `optional`)
        :return: `При успехе`: количество треков (``int``). `Если owner (пользователь или группа) не найден или треки отсутствуют`: ``0``.
        """

        if not ownerId:
            ownerId = await self._getMyId()

        trackCount = await self._req(
            "getCount",
            {
                "owner_id": ownerId,
            },
        )

        return trackCount

    get_track_count = getTrackCount


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

class GetPlaylist:
    from typing import Union

    from vkmusix.aio import async_
    from vkmusix.types import Album, Playlist

    @async_
    async def getPlaylist(self, playlistId: int, ownerId: int = None, includeTracks: bool = False) -> Union[Playlist, Album, None]:
        """
        Получает информацию о плейлисте или альбоме.

        `Пример использования`:

        playlist = client.getPlaylist(
            ownerId=-2000201020,
            playlistId=19201020,
            includeTracks=True,
        )

        print(playlist)

        :param playlistId: идентификатор плейлиста или альбома. (``int``)
        :param ownerId: идентификатор владельца плейлиста или альбома (пользователь или группа). (``int``, `optional`)
        :param includeTracks: флаг, указывающий, небходимо ли также получить треки. (``bool``, `optional`)
        :return: `При успехе`: информация о плейлисте или альбоме (``Union[types.Playlist, types.Album]``). `Если плейлист или альбом не найден`: ``None``.
        """

        from asyncio import gather

        from vkmusix.types import Playlist

        if not ownerId:
            ownerId = await self._getMyId()

        tasks = [self._req(
            "getPlaylistById",
            {
                "owner_id": ownerId,
                "playlist_id": playlistId,
            },
        )]

        if includeTracks:
            tasks.append(self.getPlaylistTracks(playlistId, ownerId))

        responses = await gather(*tasks)

        playlist = responses[0]
        if not playlist:
            return

        if includeTracks:
            playlist["tracks"] = responses[1]

        return self._finalizeResponse(playlist, Playlist)

    get_playlist = getPlaylist


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

from .base import Base

class Playlist(Base):
    """
    Класс, представляющий плейлист.

    Атрибуты:
        title (str): название плейлиста.

        subtitle (str, optional): подзаголовок плейлиста, во ВКонтакте отображается серым цветом справа от названия.

        fullTitle (str): полное название плейлиста в формате {title} ({subtitle}).

        description (str, optional): описание плейлиста.

        streams (int): количество прослушиваний плейлиста.

        saves (int): количество сохранений плейлиста пользователями или группами в свою музыку.

        createdAt (datetime): дата и время создания плейлиста.

        updatedAt (datetime): дата и время обновления плейлиста.

        original (types.Playlist, optional): оригинальный плейлист, если это не созданный, а сохранённый пользователем или группой в свою музыку.

        trackCount (int): количество треков в плейлисте.

        tracks (list[types.Track, optional): треки плейлиста. Доступны при получении через client.getPlaylist(includeTracks=True).

        own (bool, optional): флаг, указывающий, является ли плейлист созданным этим пользователем или группой. Доступен только при получении через client.getPlaylists() или client.getAllPlaylists().

        ownerId (int): идентификатор владельца плейлиста (пользователь или группа).

        playlistId (int): идентификатор плейлиста.

        id (str): полный идентификатор плейлиста в формате {ownerId}_{playlistId}.

        url (str): ссылка на плейлист в формате https://vk.com/music/playlist/{id}.

        raw (dict): необработанные данные, полученные от ВКонтакте.
    """

    from typing import Union, List, Tuple

    from vkmusix.aio import async_
    from vkmusix.types.track import Track

    def __init__(self, playlist: dict, isOwn: bool = None, client: "Client" = None) -> None:
        import html

        from vkmusix.config import VK
        from vkmusix.utils import unixToDatetime

        super().__init__(client)

        title = playlist.get("title")
        self.title = html.unescape(title) if title else None

        subtitle = playlist.get("subtitle")
        self.subtitle = html.unescape(subtitle.replace("\n", " ")) if subtitle else None

        self.fullTitle = f"{self.title} ({self.subtitle})".replace("((", "(").replace("))", ")").replace("([", "(").replace("])", ")") if self.subtitle else self.title

        description = playlist.get("description")
        self.description = description if description else None

        self.streams = playlist.get("plays")
        self.saves = playlist.get("followers")

        self.createdAt = unixToDatetime(playlist.get("create_time"))
        self.updatedAt = unixToDatetime(playlist.get("update_time"))

        photo = playlist.get("photo")
        if not photo:
            photo = playlist.get("thumb")
        self.photo = {int(key.split("_")[1]): value[:value.rfind("&c_uniq_tag=")] for key, value in photo.items() if key.startswith("photo_")} if photo else None

        self.original = self._client._finalizeResponse(
            playlist.get("original"),
            Playlist,
        )

        tracks = playlist.get("tracks")
        self.trackCount = playlist.get("count") or (len(tracks) if tracks else None)
        self.tracks = tracks

        self.own = isOwn

        self.ownerId = playlist.get("owner_id")
        self.playlistId = playlist.get("id") or playlist.get("playlist_id")
        self.id = f"{self.ownerId}_{self.playlistId}"
        self.url = f"{VK}music/playlist/{self.id}"

        self.raw = playlist


    @async_
    async def get(self, includeTracks: bool = False) -> Union["Playlist", None]:
        """
        Получает информацию о плейлисте.

        `Пример использования`:

        playlist = playlist.get(
            includeTracks=True,
        )

        print(playlist)

        :param includeTracks: флаг, указывающий, небходимо ли также получить треки. (``bool``, `optional`)
        :return: `При успехе`: информация о плейлисте (``types.Playlist``). `Если плейлист не найден`: ``None``.
        """

        return await self._client.getPlaylist(
            self.playlistId,
            self.ownerId,
            includeTracks,
        )


    @async_
    async def getTracks(self) -> Union[List[Track], None]:
        """
        Получает треки плейлиста.

        `Пример использования`:

        tracks = playlist.getTracks()

        print(tracks)

        :return: `При успехе`: треки плейлиста (``list[types.Track]``). `Если плейлист не найден или треки отсутствуют`: ``None``.
        """

        return await self._client.getPlaylistTracks(
            self.playlistId,
            self.ownerId,
        )


    @async_
    async def add(self, groupId: int = None) -> Union["Playlist", None]:
        """
        Добавляет плейлист в музыку пользователя или группы.

        `Пример использования`:

        playlist = playlist.add()

        print(playlist)

        :param groupId: идентификатор группы, в которую необходимо добавить плейлист. (``int``, `optional`)
        :return: `При успехе`: информация о добавленном плейлисте (``types.Playlist``). `Если плейлист не найден`: ``None``.
        """

        return await self._client.addPlaylist(
            self.playlistId,
            self.ownerId,
            groupId,
        )


    @async_
    async def remove(self, groupId: int = None, validateIds: bool = True) -> bool:
        """
        Удаляет плейлист из музыки пользователя или группы.

        `Пример использования`:

        result = playlist.remove()

        print(result)

        :param groupId: идентификатор группы, из которой необходимо удалить плейлист. (``int``, `optional`)
        :param validateIds: флаг, указывающий, необходимо ли перепроверить плейлист на наличие в музыке. По умолчанию ``True``. Установите на ``False``, если вы получили плейлист через ``client.getPlaylists()`` или ``client.getAllPlaylists()``. (``bool``, `optional`)
        :return: `При успехе`: ``True``. `Если плейлист не найден или не получилось его удалить`: ``False``.
        """

        return await self._client.removePlaylist(
            self.playlistId,
            self.ownerId,
            groupId,
            validateIds,
        )


    @async_
    async def edit(self, title: str = None, description: Union[str, None] = str(), photo: Union[str, None] = str()) -> bool:
        """
        Изменяет информацию о плейлисте.

        `Пример использования`:

        result = playlist.edit(
            title="Лучшая музыка в машину!!!",
        )

        print(result)

        :param title: название плейлиста. (``str``, `optional`)
        :param description: описание плейлиста. ``None`` для удаления. (``Union[str, None]``, `optional`)
        :param photo: ссылка на фото плейлиста. ``None`` для удаления. Не для удаления не работает. (``Union[str, None]``, `optional`)
        :return: `При успехе`: ``True``. `Если информацию о плейлисте не удалось изменить`: ``False``.
        """

        return await self._client.editPlaylist(
            self.playlistId,
            title,
            description,
            photo,
            self.ownerId,
        )


    @async_
    async def copy(self, groupId: int = None, chatId: int = None, title: Union[str, None] = str(), description: Union[str, None] = str(), photo: Union[str, None] = str()) -> Union["Playlist", None]:
        """
        Копирует плейлист в музыку пользователя или группы.

        `Пример использования`:

        playlist = playlist.copy()

        print(playlist)

        :param groupId: идентификатор группы, в которую необходимо скопировать плейлист. (``int``, `optional`)
        :param chatId: идентификатор чата, к которому необходимо привязать скопированный плейлист. (``int``, `optional`)
        :param title: название плейлиста. ``None`` для удаления. (``Union[str, None]``, `optional`)
        :param description: описание плейлиста. ``None`` для удаления. (``Union[str, None]``, `optional`)
        :param photo: ссылка на фото плейлиста. ``None`` для удаления. Не для удаления не работает. (``Union[str, None]``, `optional`)
        :return: `При успехе`: информация о скопированном плейлисте (``types.Playlist``). `Если плейлист не найден`: ``None``.
        """

        return await self._client.copyPlaylist(
            self.playlistId,
            self.ownerId,
            groupId,
            chatId,
            title,
            description,
            photo,
        )


    @async_
    async def addTracks(self, ownerIds: Union[List[int], int], trackIds: Union[List[int], int]) -> List[bool]:
        """
        Добавляет треки в плейлист пользователя или группы.

        `Пример использования`:

        result = playlist.addTracks(
            ownerIds=-2001471901,
            trackIds=123471901,
        )

        print(result)

        :param ownerIds: идентификаторы владельцев треков. (``Union[list[int], int]``)
        :param trackIds: идентификаторы треков. (``Union[list[int], int]``)
        :return: Статусы добавления треков (``list[bool]``). `При успехе`: ``True``. `Если трек не удалось добавить`: ``False``.
        """

        return await self._client.add(
            ownerIds,
            trackIds,
            self.playlistId,
            self.ownerId,
        )


    @async_
    async def removeTracks(self, ownerIds: Union[List[int], int], trackIds: Union[List[int], int], validateIds: bool = True) -> List[bool]:
        """
        Удаляет треки из плейлиста пользователя или группы.

        `Пример использования`:

        result = playlist.removeTracks(
            ownerIds=-2001471901,
            trackIds=123471901,
        )

        print(result)

        :param ownerIds: идентификатор владельца треков. (``Union[list[int], int]``)
        :param trackIds: идентификаторы треков. (``Union[list[int], int]``)
        :param validateIds: флаг, указывающий, необходимо ли перепроверить треки на наличие в плейлисте. По умолчанию ``True``. Установите на ``False``, если вы получили треки через ``client.getPlaylistTracks()``. (``bool``, `optional`)
        :return: Статусы удаления треков (``list[bool]``). `При успехе`: ``True``. `Если трек не удалось удалить`: ``False``.
        """

        return await self._client.remove(
            ownerIds,
            trackIds,
            self.playlistId,
            self.ownerId,
            validateIds,
        )


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

class GetRelatedArtists:
    from typing import Union, List

    from vkmusix.aio import async_
    from vkmusix.types import Artist

    @async_
    async def getRelatedArtists(self, artistId: int, limit: int = None, offset: int = None) -> Union[List[Artist], None]:
        """
        Получает похожих артистов.

        `Пример использования`:

        artists = client.getRelatedArtists(
            artistId=5696274288194638935,
            limit=10,
        )

        print(artists)

        :param artistId: идентификатор артиста. (``int``)
        :param limit: лимит артистов. (``int``, `optional`)
        :param offset: сколько артистов пропустить. (``int``, `optional`)
        :return: `При успехе`: похожие артисты (``list[types.Artist]``). `Если артист не найден или похожие артисты отсутствуют`: ``None``.
        """

        from vkmusix.types import Artist

        artists = await self._req(
            "getRelatedArtistsById",
            {
                "artist_id": artistId,
                "count": limit,
                "offset": offset,
            },
        )

        return self._finalizeResponse(artists.get("artists"), Artist)

    get_related_artists = getRelatedArtists


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

class Api:
    from typing import Union, List

    from vkmusix.aio import async_

    @async_
    async def api(self, method: str, params: dict = None, json: dict = None, data: Union[str, dict] = None, cookies: dict = None, headers: dict = None, files: dict = None, version: int = None, httpMethod: str = None) -> Union[dict, List[dict], bool, int, None]:
        """
        Отправляет запрос к любому методу Вконтакте API.

        `Пример использования`:

        result = client.api(
            method="wall.post",
            data={
                "message": "Привет!",
            },
            httpMethod="POST",
        )

        print(result)

        :param method: метод, к которому необходимо отправить запрос. (``str``)
        :param params: параметры в ссылке. (``dict``, `optional`)
        :param json: тело запроса в формате JSON. (``dict``, `optional`)
        :param data: данные в теле запроса. (``Union[str, dict]``, `optional`)
        :param cookies: cookie запроса. (``dict``, `optional`)
        :param headers: HTTP-заголовки запроса. (``dict``, `optional`)
        :param files: файлы для загрузки. Для POST-запросов. (``dict``, `optional`)
        :param version: версия ВКонтакте API. (``float``, `optional`)
        :param httpMethod: HTTP-метод запроса. По умолчанию GET. (``str``, `optional`)
        :return: Ответ ВКонтакте API (``Union[dict, List[dict], bool, int, None]``).
        """

        return await self._req(
            method=method,
            params=params,
            json=json,
            data=data,
            cookies=cookies,
            headers=headers,
            files=files,
            version=version,
            httpMethod=httpMethod.upper() if httpMethod and isinstance(httpMethod, str) else "GET",
        )


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

class GetLyrics:
    from typing import Union

    from vkmusix.aio import async_

    @async_
    async def getLyrics(self, ownerId: int, trackId: int) -> Union[str, None]:
        """
        Получает текст трека.

        `Пример использования`:

        lyrics = client.getLyrics(
            ownerId=-2001471901,
            trackId=123471901,
        )

        print(lyrics)

        :param ownerId: идентификатор владельца трека. (``int``)
        :param trackId: идентификатор трека. (``int``)
        :return: `При успехе`: текст трека (``str``). `Если трек не найден или текст отсутствует`: ``None``.
        """

        lyrics = await self._req(
            "getLyrics",
            {
                "audio_id": f"{ownerId}_{trackId}",
            },
        )

        if not lyrics:
            return

        lyrics = lyrics.get("lyrics")
        timestamps = lyrics.get("timestamps")

        return "\n".join([line.get("line") for line in timestamps if line.get("line") is not None] if timestamps else lyrics.get("text"))

    get_lyrics = getLyrics


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

from .base import Base

class Artist(Base):
    """
    Класс, представляющий артиста.

    Атрибуты:
        nickname (str): псевдоним артиста.

        photo (dict, optional): словарь с ссылками на различные размеры фотографии артиста, отсортированные по возрастанию.

        albums (list[types.Album], optional): альбомы артиста. Доступны при получении через client.getArtist(includeAlbums=True).

        tracks (list[types.Track], optional): треки артиста. Доступны при получении через client.getArtist(includeTracks=True).

        domain (str, optional): уникальное имя артиста, использующееся в ссылке.

        id (int, optional): идентификатор артиста. Недоступен для Various Artists.

        url (str, optional): ссылка на артиста в формате https://vk.com/artist/{domain or id}. Недоступна для Various Artists.

        raw (dict): необработанные данные, полученные от ВКонтакте.
    """

    from typing import Union, List

    from vkmusix.aio import async_
    from vkmusix.types.album import Album
    from vkmusix.types.track import Track

    def __init__(self, artist: dict, client: "Client" = None) -> None:
        import re
        import html

        from vkmusix.config import VK

        super().__init__(client)

        nickname = artist.get("name")
        self.nickname = html.unescape(nickname) if nickname else None

        photo = artist.get("photo")
        if photo:
            photoDict = {int(photo.get("width")): re.sub(r"\.(j|jp|jpg)$", str(), photo.get("url")[:photo.get("url").rfind("&c_uniq_tag=")][:photo.get("url").rfind("&type=")]) for photo in photo}
            self.photo = dict(sorted(photoDict.items(), key=lambda item: (int(item[0]))))

        else:
            photo = artist.get("photos")
            self.photo = dict(sorted({int(photo.get("width")): photo.get("url") for photo in photo[0].get("photo")}.items(), key=lambda item: (int(item[0])))) if photo else None

        self.albums = artist.get("albums")
        self.tracks = artist.get("tracks")

        domain = artist.get("domain")
        id = artist.get("id") or artist.get("artist_id")
        self.domain = domain if domain != id else None
        self.id = id
        self.url = f"{VK}artist/{self.domain or self.id}" if self.domain or self.id else None

        self.raw = artist


    @async_
    async def get(self, includeAlbums: bool = False, includeTracks: bool = False) -> Union["Artist", None]:
        """
        Получает информацию об артисте.

        `Пример использования`:

        artist = artist.get(
            includeAlbums=True,
            includeTracks=True,
        )

        print(artist)

        :param includeAlbums: флаг, указывающий, небходимо ли также получить альбомы. (``bool``, `optional`)
        :param includeTracks: флаг, указывающий, небходимо ли также получить треки. (``bool``, `optional`)
        :return: `При успехе`: информация об артисте (``types.Artist``). `Если артист не найден`: ``None``.
        """

        return await self._client.getArtist(
            self.id,
            includeAlbums,
            includeTracks,
        )


    @async_
    async def getAlbums(self, limit: int = None, offset: int = None) -> Union[List[Album], None]:
        """
        Получает альбомы артиста.

        `Пример использования`:

        albums = artist.getAlbums(
            limit=10,
        )

        print(albums)

        :param limit: лимит альбомов. (``int``, `optional`)
        :param offset: сколько альбомов пропустить. (``int``, `optional`)
        :return: `При успехе`: альбомы артиста (``list[types.Album]``). `Если артист не найден или альбомы отсутствуют`: ``None``.
        """

        return await self._client.getArtistAlbums(
            self.id,
            limit,
            offset,
        )


    @async_
    async def getTracks(self, limit: int = None, offset: int = None) -> Union[List[Track], None]:
        """
        Получает треки артиста.

        `Пример использования`:

        tracks = artist.getTracks(
            limit=10,
        )

        print(tracks)

        :param limit: лимит треков. (``int``, `optional`)
        :param offset: сколько треков пропустить. (``int``, `optional`)
        :return: `При успехе`: треки артиста (``list[types.Track]``). `Если артист не найден или треки отсутствуют`: ``None``.
        """

        return await self._client.getArtistTracks(
            self.id,
            limit,
            offset,
        )


    @async_
    async def getRelated(self, limit: int = None, offset: int = None) -> Union[List["Artist"], None]:
        """
        Получает похожих артистов.

        `Пример использования`:

        artists = artist.getRelated(
            limit=10,
        )

        print(artists)

        :param limit: лимит артистов. (``int``, `optional`)
        :param offset: сколько артистов пропустить. (``int``, `optional`)
        :return: `При успехе`: похожие артисты (``list[types.Artist]``). `Если артист не найден или похожие артисты отсутствуют`: ``None``.
        """

        return await self._client.getRelatedArtists(
            self.id,
            limit,
            offset,
        )


    @async_
    async def follow(self) -> bool:
        """
        Подписывается на обновления музыки артиста.

        `Пример использования`:

        result = artist.follow()

        print(result)

        :return: `При успехе`: ``True``. `Если артист не найден`: ``False``.
        """

        return await self._client.followArtist(self.id)


    @async_
    async def unfollow(self) -> bool:
        """
        Отписывается от обновлений музыки артиста.

        `Пример использования`:

        result = artist.unfollow()

        print(result)

        :return: `При успехе`: ``True``. `Если артист не найден`: ``False``.
        """

        return await self._client.unfollowArtist(self.id)


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

class _EditPlaylistPhoto:
    from typing import Union

    async def _editPlaylistPhoto(self, playlistId: int, photo: Union[str, None] = str(), groupId: int = None) -> bool:
        if photo == str():
            return False

        if not groupId:
            groupId = await self._getMyId()

        params = {
            "playlist_id": playlistId,
            "owner_id": groupId,
        }

        if photo is None:
            method = "deletePlaylistCoverPhoto"

        else:
            raise NotImplementedError

            method = "setPlaylistCoverPhoto"
            params = {
                **params,
                **{
                    "photo": photo,
                },
            }

        response = await self._req(method, params)

        return bool(response)


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

class _GetTracks:
    from typing import Union, List

    from vkmusix.types import Track

    async def _getTracks(self, tracks: str) -> Union[List[Track], None]:
        from vkmusix.types import Track

        if not tracks or "<title>Музыка ВКонтакте: слушать песни и музыку онлайн бесплатно | ВКонтакте</title>" in tracks:
            return

        tracks = tracks.replace("\\/", "/").replace("false", "False").replace("true", "True").replace("null", "None")

        try:
            tracks = eval(tracks[tracks.rfind("[["): tracks.rfind("]]") + 2])

        except SyntaxError:
            self._raiseError("accessDenied")

        if not isinstance(tracks, list):
            return

        for index, track in enumerate(tracks):
            if len(track) < 20:
                tracks[index] = None
                continue

            album = track[19]
            if album:
                album.append(dict(zip(["photo_160", "photo_300"], track[14].split(",")[::-1])))

            tracks[index] = self._finalizeResponse(
                {
                    "owner_id": track[1],
                    "track_id": track[0],
                    "title": track[3],
                    "subtitle": track[16],
                    "artist": track[4],
                    "main_artists": track[17],
                    "featured_artists": track[18],
                    "duration": track[5],
                    "album": {
                        **({
                            "photo": album[3]
                        } if len(album) == 4 else dict()),
                        "owner_id": album[0],
                        "album_id": album[1],
                    } if album else None,
                    "release_audio_id": track[-2],
                },
                Track,
            )

        tracks = [track for track in tracks if track]

        if not tracks:
            return

        return tracks


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

class SetBroadcast:
    from typing import Union, List

    from vkmusix.aio import async_

    @async_
    async def setBroadcast(self, ownerId: int = None, trackId: int = None, groupIds: Union[List[int], int] = None) -> bool:
        """
        Транслирует трек в статус owner'а (пользователь или группа). Также останавливает трансляцию в статус.

        `Пример использования для трансляции трека в статус залогиненного пользователя`:

        result = client.setBroadcast(
            ownerId=-2001471901,
            trackId=123471901,
        )

        print(result)

        `Пример использования для остановки трансляции трека в статус залогиненного пользователя`:

        result = client.setBroadcast()

        print(result)

        `Пример использования для трансляции трека в статус группы`:

        result = client.setBroadcast(
            ownerId=-2001471901,
            trackId=123471901,
            groupIds=1,
        )

        print(result)

        `Пример использования для остановки трансляции трека в статус группы`:

        result = client.setBroadcast(
            groupIds=1,
        )

        print(result)

        :param ownerId: идентификатор владельца трека, который необходимо транслировать в статус (пользователь или группа). Если параметр не заполнен, останавливает трансляцию в статус. (``int``, `optional`)
        :param trackId: идентификатор трека, который необходимо транслировать в статус. Если параметр не заполнен, останавливает трансляцию в статус. (``int``, `optional`)
        :param groupIds: идентификаторы групп, трансляцию в статус которых необходимо начать или остановить. По умолчанию залогиненный пользователь. (``Union[list[int], int]``, `optional`)
        :return: `При успехе`: ``True``. `В противном случае`: ``False``.
        """

        response = await self._req(
            "setBroadcast",
            {
                **(
                    {
                        "audio": f"{ownerId}_{trackId}",
                    }
                    if ownerId and trackId else dict()
                ),
                **(
                    {
                        "target_ids": groupIds,
                    }
                    if groupIds else dict()
                )
            },
        )

        return bool(response)

    set_broadcast = setBroadcast


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

from .error import Error
from .unknown import Unknown

from .sessionClosed import SessionClosed
from .sessionAlreadyClosed import SessionAlreadyClosed
from .sessionAlreadyOpened import SessionAlreadyOpened

from .vkInvalidToken import VKInvalidToken
from .vkCookieFileNotFound import VKCookieFileNotFound
from .vkInvalidCookie import VKInvalidCookie
from .vkUnsuccessfulLoginAttempt import VKUnsuccessfulLoginAttempt

from .ruCaptchaInvalidKey import RuCaptchaInvalidKey
from .ruCaptchaZeroBalance import RuCaptchaZeroBalance
from .ruCaptchaBannedIP import RuCaptchaBannedIP
from .ruCaptchaBannedAccount import RuCaptchaBannedAccount

from .invalidMethod import InvalidMethod
from .accessDenied import AccessDenied
from .accessDeniedWithoutCookie import AccessDeniedWithoutCookie

from .userWasDeletedOrBanned import UserWasDeletedOrBanned
from .trackRestorationTimeEnded import TrackRestorationTimeEnded

from .notFound import NotFound
from .chatNotFound import ChatNotFound

from .noneQuery import NoneQuery

from .ownerIdsAndTrackIdsTypeDifferent import OwnerIdsAndTrackIdsTypeDifferent
from .ownerIdsAndTrackIdsLenDifferent import OwnerIdsAndTrackIdsLenDifferent

from .trackReorderNeedsBeforeOrAfterArgument import TrackReorderNeedsBeforeOrAfterArgument
from .trackReorderNeedsOnlyBeforeOrAfterNotBoth import TrackReorderNeedsOnlyBeforeOrAfterNotBoth

from .mp3FileNotFound import MP3FileNotFound
from .mp3FileTooBig import MP3FileTooBig

from .tooHighRequestSendingRate import TooHighRequestSendingRate

from .invalidProxyType import InvalidProxyType
from .invalidProxyDict import InvalidProxyDict


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

class Reorder:
    from vkmusix.aio import async_

    @async_
    async def reorder(self, trackId: int, beforeTrackId: int = None, afterTrackId: int = None) -> bool:
        """
        Изменяет порядок трека в музыке пользователя. Должен быть заполнен один из параметров на выбор: ``beforeTrackId`` или ``afterTrackId``.

        `Пример использования для перемещения на место перед определённым треком`:

        result = client.reorder(
            trackId=123471901,
            beforeTrackId=123471901,
        )

        print(result)

        `Пример использования для перемещения на место после определённого трека`:

        result = client.reorder(
            trackId=123471901,
            afterTrackId=123471901,
        )

        print(result)

        :return: ``True``.
        """

        if not any((beforeTrackId, afterTrackId)):
            self._raiseError("trackReorderNeedsBeforeOrAfterArgument")

        if all((beforeTrackId, afterTrackId)):
            self._raiseError("trackReorderNeedsOnlyBeforeOrAfterNotBoth")

        response = await self._req(
            "reorder",
            {
                "audio_id": trackId,
                **(
                    {
                        "before": beforeTrackId,
                    }
                    if beforeTrackId else
                    {
                        "after": afterTrackId,
                    }
                ),
            }
        )

        return bool(response)


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

class SearchAlbums:
    from typing import Union, List

    from vkmusix.aio import async_
    from vkmusix.types import Album

    @async_
    async def searchAlbums(self, query: str, limit: int = None, offset: int = None) -> Union[List[Album], None]:
        """
        Ищет альбомы.

        `Пример использования`:

        albums = client.searchAlbums(
            query="Маленький ярче — ANDERFUL ELF",
            limit=10,
        )

        print(albums)

        :param query: поисковой запрос. (``str``)
        :param limit: лимит альбомов. (``int``, `optional`)
        :param offset: сколько альбомов пропустить. (``int``, `optional`)
        :return: `При успехе`: найденные альбомы (``list[types.Album]``). `Если альбомы не найдены`: ``None``.
        """

        from vkmusix.types import Album

        return await self._search(
            "searchAlbums",
            (query, limit, offset),
            Album,
        )

    search_albums = searchAlbums


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

class GetSearchTrends:
    from typing import Union, List

    from vkmusix.aio import async_

    @async_
    async def getSearchTrends(self, limit: int = None, offset: int = None) -> Union[List[str], None]:
        """
        Получает популярные поисковые запросы.

        `Пример использования`:

        searches = client.getSearchTrends(
            limit=10,
        )

        print(searches)

        :param limit: лимит поисковых запросов. (``int``, `optional`)
        :param offset: сколько поисковых запросов пропустить. (``int``, `optional`)
        :return: `При успехе`: поисковые запросы (``list[str]``). `Если поисковые запросы отсутствуют`: ``None``.
        """

        searchTrends = await self._req(
            "getSearchTrends",
            {
                "count": limit,
                "offset": offset,
            },
        )

        searchTrends = [
            item.get("name")
            for item in searchTrends.get("items")
        ]

        return searchTrends if searchTrends else None

    get_search_trends = getSearchTrends


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

class CopyPlaylist:
    from typing import Union

    from vkmusix.aio import async_
    from vkmusix.types import Playlist

    @async_
    async def copyPlaylist(self, playlistId: int, ownerId: int = None, groupId: int = None, chatId: int = None, title: Union[str, None] = str(), description: Union[str, None] = str(), photo: Union[str, None] = str()) -> Union[Playlist, None]:
        """
        Копирует плейлист или альбом в музыку пользователя или группы.

        `Пример использования`:

        playlist = client.copyPlaylist(
            ownerId=-2000201020,
            playlistId=19201020,
        )

        print(playlist)

        :param playlistId: идентификатор плейлиста или альбома. (``int``)
        :param ownerId: идентификатор владельца плейлиста или альбома (пользователь или группа). (``int``, `optional`)
        :param groupId: идентификатор группы, в которую необходимо скопировать плейлист или альбом. (``int``, `optional`)
        :param chatId: идентификатор чата, к которому необходимо привязать скопированный плейлист или альбом. (``int``, `optional`)
        :param title: название плейлиста. ``None`` для удаления. (``Union[str, None]``, `optional`)
        :param description: описание плейлиста. ``None`` для удаления. (``Union[str, None]``, `optional`)
        :param photo: ссылка на фото плейлиста. ``None`` для удаления. Не для удаления не работает. (``Union[str, None]``, `optional`)
        :return: `При успехе`: информация о скопированном плейлисте или альбоме (``Union[types.Playlist, types.Album]``). `Если плейлист или альбом не найден`: ``None``.
        """

        from datetime import datetime

        if not ownerId:
            ownerId = await self._getMyId()

        playlist = await self.getPlaylist(playlistId, ownerId, True)

        title_ = playlist.title
        description_ = playlist.description
        photo_ = playlist.photo

        if title != str():
            title_ = title if title is not None else datetime.utcnow().strftime("%d.%m.%Y / %H:%M:%S")

        if description != str():
            description_ = description

        if photo != str():
            photo_ = None if True else photo

        elif photo_:
            _, photo_ = photo_.popitem()

        newPlaylist = await self.createPlaylist(
            title_,
            description_,
            photo_,
            groupId,
            chatId,
        )

        tracks = playlist.tracks
        if tracks:
            ownerIds, trackIds = zip(*[(track.ownerId, track.trackId) for track in tracks[::-1]])
            await newPlaylist.addTracks(list(ownerIds), list(trackIds))

        return newPlaylist

    copy_playlist = copyPlaylist


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

class GetNew:
    from typing import Union, List

    from vkmusix.aio import async_
    from vkmusix.types import Track

    @async_
    async def getNew(self) -> Union[List[Track], None]:
        """
        Получает треки, вышедшие недавно.

        `Пример использования`:

        tracks = client.getNew()

        print(tracks)

        :return: `При успехе`: треки (``list[types.Track]``). `Если треки отсутствуют`: ``None``.
        """

        from vkmusix.config import playlistsOwnerId

        tracks = await self.getPlaylistTracks(2, playlistsOwnerId)

        return tracks

    get_new = getNew


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

import os
import platform

from typing import Union

from datetime import datetime

def unixToDatetime(seconds: Union[int, None]) -> Union[datetime, None]:
    if not seconds:
        return

    return datetime.utcfromtimestamp(seconds)


def fileExistsCaseInsensitive(filename: str) -> Union[str, None]:
    directory, filename = os.path.split(filename)
    if not directory:
        directory = "."

    for root, _, files in os.walk(directory):
        for name in files:
            if name.lower() == filename.lower():
                return os.path.join(root, name)

    return


def checkFile(filename: str) -> Union[str, None]:
    if os.path.isfile(filename):
        return filename

    system = platform.system()
    if system not in ["Linux", "Darwin"]:
        return

    filename = fileExistsCaseInsensitive(filename)
    if not filename:
        return


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

class GetRecommendations:
    from typing import Union, List

    from vkmusix.aio import async_
    from vkmusix.types import Track

    @async_
    async def getRecommendations(self, limit: int = None, offset: int = None, ownerId: int = None, trackId: int = None) -> Union[List[Track], None]:
        """
        Получает пользовательские рекомендации, или рекомендации по треку, если заполнены параметры ``ownerId`` и ``trackId``.

        `Пример использования для пользовательских рекомендаций`:

        tracks = client.getRecommendations(
            limit=10,
        )

        print(tracks)

        `Пример использования для рекомендаций по треку`:

        tracks = client.getRecommendations(
            limit=10,
            ownerId=-2001471901,
            trackId=123471901,
        )

        print(tracks)

        :param limit: лимит треков. Для пользовательских рекомендаций минимально 10. (``int``, `optional`)
        :param offset: сколько треков пропустить. (``int``, `optional`)
        :param ownerId: идентификатор владельца трека. (``int``, `optional`)
        :param trackId: идентификатор трека. (``int``, `optional`)
        :return: `При успехе`: рекомендации (``list[types.Track]``). `Если рекомендации отсутствуют или трек не найден`: ``None``.
        """

        from vkmusix.types import Track

        if not all((ownerId, trackId)) and limit < 10:
            limit = 10

        tracks = await self._req(
            "getRecommendations",
            {
                "count": limit,
                "offset": offset,
                **(
                    {
                        "target_audio": f"{ownerId}_{trackId}"
                    }
                    if all((ownerId, trackId)) else dict(),
                ),
            },
        )

        return self._finalizeResponse(tracks.get("items"), Track)

    get_recommendations = getRecommendations


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

from .base import Base

class Track(Base):
    """
    Класс, представляющий трек.

    Атрибуты:
        title (str): название трека.

        subtitle (str, optional): подзаголовок трека, во ВКонтакте отображается серым цветом справа от названия.

        fullTitle (str): полное название трека в формате {title} ({subtitle}).

        artist (str): все артисты трека в виде строки.

        artists (list[types.Artist], optional): основные артисты трека. Доступно только для официально загруженных треков.

        featuredArtists (list[types.Artist], optional): приглашённые артисты трека. Доступно только для оффициально загруженных треков.

        duration (int): длительность трека в секундах.

        genre (types.Genre, optional): жанр трека.

        lyrics (str, optional): текст трека.

        hasLyrics (bool, optional): флаг, указывающий, имеет ли трек текст. Отсутствует, если lyrics не None.

        uploadedAt (datetime): дата и время загрузки трека (не релиза).

        fileUrl (str, optional) — ссылка на файл трека в формате .M3U8. Отсутствует, если трек доступен только с подпиской, а залогиненный пользователь её не имеет.

        album (types.Album, optional): альбом, на котором присутствует этот трек. В некоторых случаях может быть доступно и не для оффициально загруженных треков.

        explicit (bool, optional): флаг, указывающий, есть ли в треке ненормативная лексика. Доступно только для оффициально загруженных треков.

        licensed (bool, optional): флаг, указывающий, ???

        focus (bool, optional): флаг, указывающий, является ли трек фокус-треком на альбоме.

        shortsAllowed (bool, optional): флаг, указывающий, доступен ли этот трек для использования в ВК Клипах.

        storiesAllowed (bool, optional): флаг, указывающий, доступен ли этот трек для использования в историях.

        releaseTrack (types.Track, optional): официально загруженный трек, который ВКонтакте считает максимально похожим на данный. Может быть этим же треком.

        ownerId (int): идентификатор владельца трека (пользователь или группа).

        trackId (int): идентификатор трека.

        id (str): полный идентификатор трека в формате {ownerId}_{trackId}.

        url (str): ссылка на трек в формате https://vk.com/audio{id}

        raw (dict): необработанные данные, полученные от ВКонтакте.
    """

    from typing import Union, List

    from vkmusix.aio import async_
    from vkmusix.enums import Extension

    def __init__(self, track: dict, releaseTrack: bool = None, client: "Client" = None) -> None:
        import html

        from vkmusix.config import VK
        from vkmusix.utils import unixToDatetime

        from vkmusix.types.artist import Artist
        from vkmusix.types.album import Album
        from vkmusix.types.genre import Genre

        super().__init__(client)

        title = track.get("title")
        self.title = html.unescape(title) if title else None

        subtitle = track.get("subtitle")
        self.subtitle = html.unescape(subtitle.replace("\n", " ")) if subtitle else None

        self.fullTitle = f"{self.title} ({self.subtitle})".replace("((", "(").replace("))", ")").replace("([", "(").replace("])", ")") if self.subtitle else self.title

        artist = track.get("artist")
        self.artist = html.unescape(artist) if artist else None

        self.artists = self._client._finalizeResponse(
            track.get("main_artists"),
            Artist,
        )

        self.featuredArtists = self._client._finalizeResponse(
            track.get("featured_artists"),
            Artist,
        )

        self.duration = track.get("duration")

        genreId = track.get("genre_id")
        self.genre = Genre(
            genreId=genreId,
            client=self._client,
        ) if genreId else None

        self.lyrics = track.get("lyrics")
        self.hasLyrics = track.get("has_lyrics") if not self.lyrics else None

        self.uploadedAt = unixToDatetime(track.get("date"))

        self.fileUrl = track.get("url") or None

        self.album = self._client._finalizeResponse(
            track.get("album"),
            Album,
        )

        self.explicit = track.get("is_excplicit")

        self.licensed = track.get("is_licensed")
        self.focus = track.get("is_focus_track")

        self.shortsAllowed = track.get("short_videos_allowed")
        self.storiesAllowed = track.get("stories_allowed")

        self.releaseTrack = None
        if not releaseTrack:
            releaseTrackId = track.get("release_audio_id")

            if releaseTrackId:
                releaseTrackOwnerId, releaseTrackTrackId = tuple(map(int, releaseTrackId.split("_")))
                self.releaseTrack = Track(
                    {
                        "owner_id": releaseTrackOwnerId,
                        "track_id": releaseTrackTrackId,
                    },
                    True,
                    client=self._client,
                )

        self.ownerId = track.get("owner_id")
        self.trackId = track.get("id") or track.get("track_id")
        self.id = f"{self.ownerId}_{self.trackId}"
        self.url = f"{VK}audio{self.id}"

        self.raw = track


    @async_
    async def get(self, includeLyrics: bool = False) -> Union["Track", None]:
        """
        Получает информацию о треке.

        `Пример использования`:

        track = track.get(
            includeLyrics=True,
        )

        print(track)

        :param includeLyrics: флаг, указывающий, небходимо ли также получить текст. (``bool``, `optional`)
        :return: `При успехе`: информация о треке (``types.Track``). `Если трек не найден`: ``None``.
        """

        return await self._client.get(
            self.ownerId,
            self.trackId,
            includeLyrics,
        )


    @async_
    async def getLyrics(self) -> Union[str, None]:
        """
        Получает текст трека.

        `Пример использования`:

        lyrics = track.getLyrics()

        print(lyrics)

        :return: `При успехе`: текст трека (``str``). `Если трек не найден или текст отсутствует`: ``None``.
        """

        return await self._client.getLyrics(
            self.ownerId,
            self.trackId,
        )


    @async_
    async def download(self, filename: str = None, directory: str = None, extension: Extension = None, metadata: bool = False) -> Union[str, None]:
        """
        Скачивает трек.

        `Пример использования`:

        from vkmusix.enums import Extension

        path = track.download(
            extension=Extension.OPUS,
            metadata=True,
        )

        print(path)

        :param filename: имя файла с треком. По умолчанию ``{artist} — {fullTitle}``. Поддерживаемые переменные для динамического имени: ``artist``, ``title``, ``subtitle``, ``fullTitle``, ``album``. Пример динамического имени файла: ``{artist} - {title} ({album})``. (``str``, `optional`)
        :param directory: путь к директории, в которую загрузить трек. (``str``, `optional`)
        :param extension: расширение файла с треком. По умолчанию ``Extension.MP3``. (``enums.Extension``, `optional`)
        :param metadata: флаг, указывающий, необходимо ли добавить метаданные (артист, название, альбом, обложка) к файлу с треком. По умолчанию ``False``. Игнорируется, если параметр ``extension`` равен ``Extension.TS``. (``bool``, `optional`)
        :return: `При успехе`: полный путь к загруженному файлу (``str``). `Если трек не найден или недоступен для загрузки`: ``None``.
        """

        return await self._client.download(
            filename=filename,
            directory=directory,
            extension=extension,
            metadata=metadata,
            track=self,
        )


    @async_
    async def getRecommendations(self, limit: int = None, offset: int = None) -> Union[List["Track"], None]:
        """
        Получает рекомендации по треку.

        `Пример использования`:

        tracks = track.getRecommendations(
            limit=10,
        )

        print(tracks)

        :param limit: лимит треков. (``int``, `optional`)
        :param offset: сколько треков пропустить. (``int``, `optional`)
        :return: `При успехе`: рекомендации (``list[types.Track]``). `Если рекомендации отсутствуют или трек не найден`: ``None``.
        """

        return await self._client.getRecommendations(
            limit,
            offset,
            self.ownerId,
            self.trackId,
        )


    @async_
    async def add(self, playlistId: int = None, groupId: int = None) -> bool:
        """
        Добавляет трек в музыку или плейлист пользователя или группы.

        `Пример использования`:

        result = track.add()

        print(result)

        :param playlistId: идентификатор плейлиста, в который необходимо добавить трек. (``int``, `optional`)
        :param groupId: идентификатор группы, в музыку или плейлист которой необходимо добавить трек. (``int``, `optional`)
        :return: `При успехе`: ``True``. `Если трек не удалось добавить`: ``False``.
        """

        return (await self._client.add(
            self.ownerId,
            self.trackId,
            playlistId,
            groupId,
        ))[0]


    @async_
    async def remove(self, playlistId: int = None, groupId: int = None, validateIds: bool = True) -> bool:
        """
        Удаляет трек из музыки или плейлиста пользователя или группы.

        `Пример использования`:

        result = track.remove()

        print(result)

        :param playlistId: идентификатор плейлиста, из которого необходимо удалить трек. Метод не работает для плейлистов, привязанных к чату (``int``, `optional`)
        :param groupId: идентификатор группы, из музыки или плейлиста которой необходимо удалить трек. (``int``, `optional`)
        :param validateIds: флаг, указывающий, необходимо ли перепроверить трек на наличие в музыке или плейлисте. По умолчанию ``True``. Установите на ``False``, если вы получили трек через ``client.getTracks()`` (при удалении из музыки) или ``client.getPlaylistTracks()`` (при удалении из плейлиста). (``bool``, `optional`)
        :return: Статусы удаления треков (``list[bool]``). `При успехе`: ``True``. `Если трек не удалось удалить`: ``False``.
        """

        return (await self._client.remove(
            self.ownerId,
            self.trackId,
            playlistId,
            groupId,
            validateIds,
        ))[0]


    @async_
    async def edit(self, title: str = None, artist: str = None, lyrics: str = None, genreId: int = None, removeFromSearchResults: bool = None) -> bool:
        """
        Изменяет информацию о треке.

        `Пример использования`:

        result = track.edit(
            filename="Маленький ярче — LARILARI",
            title="LARILARI",
            artist="Маленький ярче",
            genreId=21,
            removeFromSearchResults=True,
        )

        print(result)

        :param title: название трека. По умолчанию берётся из метаданных файла. (``str``, `optional`)
        :param artist: артисты трека. По умолчанию берётся из метаданных файла. (``str``, `optional`)
        :param lyrics: текст трека. (``str``, `optional`)
        :param genreId: идентификатор жанра трека. (``int``, `optional`)
        :param removeFromSearchResults: флаг, указывающий, необходимо ли исключить трек из поиска. По умолчанию ``False``. (``bool``, `optional`)
        :return: `При успехе`: ``True``. `Если информацию о треке не удалось изменить`: ``False``.
        """

        return await self._client.edit(
            self.ownerId,
            self.trackId,
            title,
            artist,
            lyrics,
            genreId,
            removeFromSearchResults,
        )


    @async_
    async def restore(self) -> bool:
        """
        Восстанавливает удалённый трек.

        `Пример использования`:

        result = track.restore()

        print(result)

        :return: `При успехе`: ``True``. `Если трек не удалось восстановить`: ``False``.
        """

        return await self._client.restore(
            self.ownerId,
            self.trackId,
        )


    @async_
    async def reorder(self, beforeTrackId: int = None, afterTrackId: int = None) -> bool:
        """
        Изменяет порядок трека в музыке пользователя. Должен быть заполнен один из параметров на выбор: ``beforeTrackId`` или ``afterTrackId``.

        `Пример использования для перемещения на место перед определённым треком`:

        result = track.reorder(
            beforeTrackId=123471901,
        )

        print(result)

        `Пример использования для перемещения на место после определённого трека`:

        result = track.reorder(
            afterTrackId=123471901,
        )

        print(result)

        :return: ``True``.
        """

        return await self._client.reorder(
            self.trackId,
            beforeTrackId,
            afterTrackId,
        )


    @async_
    async def setBroadcast(self, groupIds: Union[List[int], int] = None) -> bool:
        """
        Транслирует трек в статус owner'а (пользователь или группа).

        `Пример использования для трансляции трека в статус залогиненного пользователя`:

        result = track.setBroadcast()

        print(result)

        `Пример использования для трансляции трека в статус группы`:

        result = track.setBroadcast(
            groupIds=1,
        )

        print(result)

        :param groupIds: идентификаторы групп, трансляцию в статус которых необходимо начать. По умолчанию залогиненный пользователь. (``Union[list[int], int]``, `optional`)
        :return: `При успехе`: ``True``. `В противном случае`: ``False``.
        """

        return await self._client.setBroadcast(
            self.ownerId,
            self.trackId,
            groupIds,
        )


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

class Upload:
    from typing import Union

    from vkmusix.aio import async_
    from vkmusix.types import Track

    @async_
    async def upload(self, filename: str, title: str = None, artist: str = None, lyrics: str = None, genreId: int = None, removeFromSearchResults: bool = None, playlistId: int = None, groupId: int = None) -> Union[Track, None]:
        """
        Загружает новый трек во ВКонтакте.

        `Пример использования`:

        track = client.upload(
            filename="Маленький ярче — LARILARI",
            title="LARILARI",
            artist="Маленький ярче",
            genreId=21,
            removeFromSearchResults=True,
        )

        print(track)

        :param filename: путь к .MP3 (обязательно) файлу. (``str``)
        :param title: название трека. По умолчанию берётся из метаданных файла. (``str``, `optional`)
        :param artist: артисты трека. По умолчанию берётся из метаданных файла. (``str``, `optional`)
        :param lyrics: текст трека. (``str``, `optional`)
        :param genreId: идентификатор жанра трека. (``int``, `optional`)
        :param removeFromSearchResults: флаг, указывающий, необходимо ли исключить трек из поиска. По умолчанию ``False``. (``bool``, `optional`)
        :param playlistId: идентификатор плейлиста, в который необходимо добавить трек после загрузки. (``int``, `optional`)
        :param groupId: идентификатор группы, в плейлист или музыку которой необходимо добавить трек после загрузки. (``int``, `optional`)
        :return: `При успехе`: информация о загруженном треке (``types.Track``). `Если трек не удалось загрузить`: ``None``.
        """

        import os.path
        import aiofiles

        from vkmusix.types import Track
        from vkmusix.utils import checkFile

        if filename.endswith(".mp3"):
            filename = filename[:-4]

        filename = checkFile(f"{filename}.mp3")
        if not filename:
            self._raiseError("MP3FileNotFound")

        fileSizeInBytes = os.path.getsize(filename)
        fileSizeInMB = fileSizeInBytes / (1024 * 1024)

        if fileSizeInMB > 200:
            self._raiseError("MP3FileTooBig")

        uploadUrl = (await self._req("getUploadServer")).get("upload_url")

        async with aiofiles.open(filename, "rb") as file:
            fileContent = await file.read()

        uploadingFileResponse = await self._client.req(
            uploadUrl,
            files={
                "file": (
                    filename,
                    fileContent,
                    "audio/mpeg",
                ),
            },
            method="POST",
        )

        server = uploadingFileResponse.get("server")
        audio = uploadingFileResponse.get("audio")
        hash = uploadingFileResponse.get("hash")

        track = await self._req(
            "save",
            {
                "server": server,
                "audio": audio,
                "hash": hash,
                "title": title,
                "artist": artist,
            },
        )

        track = self._finalizeResponse(track, Track)

        if any((lyrics, genreId, removeFromSearchResults)):
            await track.edit(lyrics=lyrics, genreId=genreId, removeFromSearchResults=removeFromSearchResults)
            track = await track.get(True)

        if any((playlistId, groupId)):
            await track.add(playlistId, groupId)
            await track.remove()

        return track


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

class Get:
    from typing import Union

    from vkmusix.aio import async_
    from vkmusix.types import Track

    @async_
    async def get(self, ownerId: int, trackId: int, includeLyrics: bool = False) -> Union[Track, None]:
        """
        Получает информацию о треке.

        `Пример использования`:

        track = client.get(
            ownerId=-2001471901,
            trackId=123471901,
            includeLyrics=True,
        )

        print(track)

        :param ownerId: идентификатор владельца трека. (``int``)
        :param trackId: идентификатор трека. (``int``)
        :param includeLyrics: флаг, указывающий, небходимо ли также получить текст. (``bool``, `optional`)
        :return: `При успехе`: информация о треке (``types.Track``). `Если трек не найден`: ``None``.
        """

        from asyncio import gather

        from vkmusix.types import Track

        id = f"{ownerId}_{trackId}"

        tasks = [self._req(
            "getById",
            {
                "audios": id,
            },
        )]

        if includeLyrics:
            tasks.append(self.getLyrics(ownerId, trackId))

        responses = await gather(*tasks)

        track = responses[0]
        if not track:
            return

        if includeLyrics:
            track["lyrics"] = responses[1]

        return self._finalizeResponse(track, Track)


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

class GetPlaylistTracks:
    from typing import Union, List

    from vkmusix.aio import async_
    from vkmusix.types import Track

    @async_
    async def getPlaylistTracks(self, playlistId: int, ownerId: int = None) -> Union[List[Track], None]:
        """
        Получает треки плейлиста или альбома.

        `Пример использования`:

        tracks = client.getPlaylistTracks(
            ownerId=-2000201020,
            albumId=19201020,
        )

        print(tracks)

        :param playlistId: идентификатор плейлиста или альбома. (``int``)
        :param ownerId: идентификатор владельца плейлиста или альбома (пользователь или группа). (``int``, `optional`)
        :return: `При успехе`: треки плейлиста или альбома (``list[types.Track]``). `Если плейлист или альбом не найден или треки отсутствуют`: ``None``.
        """

        from vkmusix.config import VK, headers
        from vkmusix.types import Track

        if not ownerId:
            ownerId = await self._getMyId()

        tracks = await self._client.req(
            f"{VK}music/playlist/{ownerId}_{playlistId}",
            headers=headers,
            responseType="response",
        )

        statusCode = tracks.status_code

        if statusCode == 404:
            return

        while statusCode == 302:
            tracks = await self._client.req(
                f'{VK}{tracks.headers.get("Location")}',
                headers=headers,
                responseType="response",
            )
            statusCode = tracks.status_code

        tracks = await self._getTracks(tracks.text)

        if not tracks:
            tracks = (await self._req(
                "getAudioIdsBySource",
                {
                    "playlist_id": f"{ownerId}_{playlistId}",
                },
                version=5.195,
            )).get("audios")

            if tracks:
                for index, id in enumerate(tracks):
                    if id.count("_") == 3:
                        id = id[:id.rfind("_")]

                    ownerId, trackId = id.split("_")[:2]
                    tracks[index] = {
                        "owner_id": int(ownerId),
                        "track_id": int(trackId),
                    }

                return self._finalizeResponse(tracks, Track)

        return tracks if tracks else None

    get_playlist_tracks = getPlaylistTracks


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

class Restore:
    from vkmusix.aio import async_

    @async_
    async def restore(self, ownerId: int, trackId: int) -> bool:
        """
        Восстанавливает удалённый трек.

        `Пример использования`:

        result = client.restore(
            ownerId=-2001471901,
            trackId=123471901,
        )

        print(result)

        :param ownerId: идентификатор владельца трека. (``int``)
        :param trackId: идентификатор трека. (``int``)
        :return: `При успехе`: ``True``. `Если трек не удалось восстановить`: ``False``.
        """

        response = await self._req(
            "restore",
            {
                "owner_id": ownerId,
                "audio_id": trackId,
            },
        )

        return bool(response)


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

from .base import Base

class Genre(Base):
    """
    Класс, представляющий жанр альбома или трека.

    Атрибуты:
        title (str): название жанра.

        id (int): идентификатор жанра. Различен для альбомов и треков.

        raw (dict, optional): необработанные данные, полученные от ВКонтакте.
    """

    def __init__(self, genre: dict = None, genreId: int = None, client: "Client" = None) -> None:
        super().__init__(client)

        if genreId:
            trackGenres = {
                1: "Рок",
                2: "Поп",
                3: "Рэп и Хип-хоп",
                4: "Расслабляющая",
                5: "House и Танцевальная",
                6: "Инструментальная",
                7: "Метал",
                8: "Дабстеп",
                10: "Drum & Bass",
                11: "Транс",
                12: "Шансон",
                13: "Этническая",
                14: "Акустическая",
                15: "Регги",
                16: "Классическая",
                17: "Инди-поп",
                18: "Другая",
                19: "Скит",
                21: "Альтернатива",
                22: "Электро-поп и Диско",
                1001: "Джаз и Блюз"
            }

            self.title = trackGenres.get(genreId, "Неизвестен")

            self.id = genreId

        else:
            self.title = genre.get("name")

            self.id = genre.get("id") or genre.get("genre_id")

        self.raw = genre


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

class SearchArtists:
    from typing import Union, List

    from vkmusix.aio import async_
    from vkmusix.types import Artist

    @async_
    async def searchArtists(self, query: str, limit: int = None, offset: int = None) -> Union[List[Artist], None]:
        """
        Ищет артистов.

        `Пример использования`:

        artists = client.searchArtists(
            query="Маленький ярче",
            limit=10,
        )

        print(artists)

        :param query: поисковой запрос. (``str``)
        :param limit: лимит артистов. (``int``, `optional`)
        :param offset: сколько артистов пропустить. (``int``, `optional`)
        :return: `При успехе`: найденные артисты (``list[types.Artist]``). `Если артисты не найдены`: ``None``.
        """

        from vkmusix.types import Artist

        return await self._search(
            "searchArtists",
            (query, limit, offset),
            Artist,
        )

    search_artists = searchArtists


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

class GetBroadcast:
    from typing import Union

    from vkmusix.aio import async_
    from vkmusix.types import Track

    @async_
    async def getBroadcast(self, ownerId: int = None) -> Union[Track, None, bool]:
        """
        Получает трек транслируемый в статус owner'а (пользователь или группа).

        `Пример использования для получения трека, транслируемого в статус залогиненного пользователя`:

        track = client.getBroadcast()

        print(track)

        `Пример использования для получения трека, транслируемого в статус любого пользователя`:

        track = client.getBroadcast(
            ownerId=1,
        )

        print(track)

        `Пример использования для получения трека, транслируемого в статус любой группы`:

        track = client.getBroadcast(
            ownerId=-1,
        )

        print(track)

        :param ownerId: идентификатор owner'а (пользователь или группа). По умолчанию залогиненный пользователь. (``int``, `optional`)
        :return: `Если трек транслируется в статус`: транслируемый трек (``types.Track``). `Если трек не транслируется в статус`: `только для залогиненного пользователя`: флаг, указывающий, транслируются ли треки в статус при проигрывании. (``bool``). `для любого owner'а (пользователь или группа)`: ``None``.
        """

        broadcast = await self._req(
            "status.get",
            (
                {
                    "user_id": ownerId,
                }
                if ownerId > 0 else
                {
                    "group_id": -ownerId,
                }
            ) if ownerId else None
        )

        audio = broadcast.get("audio")
        if audio:
            from vkmusix.types import Track

            return self._finalizeResponse(audio, Track)

        if not ownerId or ownerId == await self._getMyId():
            isBroadcastEnabled = (await self._req("getBroadcast")).get("enabled")
            return bool(isBroadcastEnabled)

    get_broadcast = getBroadcast


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

from vkmusix.types import Artist, Album, Track, Playlist, SearchResults

classes = {
    Artist: "artists",
    Album: "albums",
    Track: "audios",
    Playlist: "playlists"
}

class _Search:
    from typing import Union, Type

    async def _search(self, method: str, params: tuple, itemClass: Type[Union[Artist, Album, Track, Playlist]]) -> Union[SearchResults, None]:
        query, limit, offset = params[0], params[1], params[2]
        if not query:
            return self._raiseError("noneQuery")

        params = {
            "q": query,
            "count": limit,
            "offset": offset,
        }

        response = await self._req(method, params)

        if not isinstance(response, dict):
            return

        if isinstance(itemClass, list):
            results = dict()
            for model, key in classes.items():
                modelObjects = response.get(key)
                if modelObjects:
                    modelObjects = modelObjects.get("items")

                if not modelObjects:
                    continue

                results[model] = self._finalizeResponse(modelObjects, model)

            return SearchResults(
                artists=results.get(Artist),
                albums=results.get(Album),
                tracks=results.get(Track),
                playlists=results.get(Playlist),
            ) if results else None

        else:
            items = response.get("items")
            return self._finalizeResponse(items, itemClass)


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

class GetPlaylists:
    from typing import Union, List

    from vkmusix.aio import async_
    from vkmusix.types import Album, Playlist
    from vkmusix.enums import PlaylistType

    @async_
    async def getPlaylists(self, ownerId: int = None, limit: int = None, offset: int = None, playlistTypes: Union[List[PlaylistType], PlaylistType] = None) -> Union[List[Union[Playlist, Album]], None]:
        """
        Получает плейлисты и (или) альбомы owner'а (пользователь или группа).

        `Пример использования`:

        playlists = client.getPlaylists(
            ownerId=-1,
            limit=10,
        )

        print(playlists)

        :param ownerId: идентификатор owner'а (пользователь или группа). По умолчанию залогиненный пользователь. (``int``, `optional`)
        :param limit: лимит плейлистов и (или) альбомов. (``int``, `optional`)
        :param offset: сколько плейлистов и (или) альбомов пропустить. (``int``, `optional`)
        :param playlistTypes: нужные типы плейлистов. (``Union[list[enums.PlaylistType], enums.PlaylistType]``, `optional`)
        :return: `При успехе`: плейлисты и (или) альбомы (`list[Union[types.Playlist, types.Album]]). `Если owner (пользователь или группа) не найден или плейлисты и (или) альбомы отсутствуют`: ``None``.
        """

        from vkmusix.types import Album, Playlist
        from vkmusix.enums import PlaylistType

        if not ownerId:
            ownerId = await self._getMyId()

        if not isinstance(playlistTypes, list):
            playlistTypes = [playlistTypes]

        playlists_ = await self._req(
            "getPlaylists",
            {
                "owner_id": ownerId,
                "count": limit,
                "offset": offset,
            },
        )

        playlists = self._finalizeResponse([playlist for playlist in playlists_.get("items")], Playlist)

        if not isinstance(playlists, list):
            playlists = [playlists]

        playlists = [playlist for playlist in playlists if (isinstance(playlist, Playlist) and (PlaylistType.Own if playlist.own else PlaylistType.Foreign) in playlistTypes) or (isinstance(playlist, Album) and PlaylistType.Album in playlistTypes)]

        return playlists if playlists else None

    get_playlists = getPlaylists


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

from .get import Get
from .getLyrics import GetLyrics
from .download import Download
from .getRecommendations import GetRecommendations
from .getTracksFromFeed import GetTracksFromFeed
from .getNew import GetNew
from .getPopular import GetPopular
from .getEditorsPick import GetEditorsPick

from .upload import Upload
from .add import Add
from .remove import Remove
from .edit import Edit
from .restore import Restore

class Tracks(
    Get,
    GetLyrics,
    GetRecommendations,
    Add,
    Remove,
    Upload,
    Edit,
    Restore,
    Download,

    GetTracksFromFeed,
    GetNew,
    GetPopular,
    GetEditorsPick,
):
    pass


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

class EditPlaylist:
    from typing import Union

    from vkmusix.aio import async_

    @async_
    async def editPlaylist(self, playlistId: int, title: str = None, description: Union[str, None] = str(), photo: Union[str, None] = str(), groupId: int = None) -> bool:
        """
        Изменяет информацию о плейлисте.

        `Пример использования`:

        result = client.editPlaylist(
            title="Лучшая музыка в машину!!!",
        )

        print(result)

        :param playlistId: идентификатор плейлиста. (``int``)
        :param title: название плейлиста. (``str``, `optional`)
        :param description: описание плейлиста. ``None`` для удаления. (``Union[str, None]``, `optional`)
        :param photo: ссылка на фото плейлиста. ``None`` для удаления. Не для удаления не работает. (``Union[str, None]``, `optional`)
        :param groupId: идентификатор группы, в которой находится плейлист. (``int``, `optional`)
        :return: `При успехе`: ``True``. `Если информацию о плейлисте не удалось изменить`: ``False``.
        """

        if not any((title, description is not None, photo is not None)):
            return False

        if not groupId:
            groupId = (await self.getSelf()).get("id")

        params = {
            "playlist_id": playlistId,
            "owner_id": groupId,
            **({"title": title} if title else dict()),
            **({"description": description} if description != str() else dict())
        }

        if any(("title" in params, "description" in params)):
            response = await self._req("editPlaylist", params)
            editInfoStatus = bool(response)

        else:
            editInfoStatus = None

        if any((editInfoStatus is True, editInfoStatus is None)) and photo != str():
            editPhotoStatus = await self._editPlaylistPhoto(playlistId, photo, groupId)
            if editInfoStatus is None:
                editInfoStatus = editPhotoStatus

        return editInfoStatus

    edit_playlist = editPlaylist


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

from vkmusix.errors import Error

class OwnerIdsAndTrackIdsLenDifferent(Error, ValueError):
    def __init__(self) -> None:
        self.ru = "Длины `ownerIds` и `trackIds` не могут отличаться."
        self.en = "The lengths of `ownerIds` and `trackIds` must be the same."


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

from .base import Base

class Album(Base):
    """
    Класс, представляющий трек.

    Атрибуты:
        title (str): название альбома.

        subtitle (str, optional): подзаголовок альбома, во ВКонтакте отображается серым цветом справа от названия.

        artist (str): все артисты альбома в виде строки.

        artists (list[types.Artist]): основные артисты альбома.

        featuredArtists (list[types.Artist], optional): приглашённые артисты альбома.

        description (str, optional): описание альбома.

        releaseYear (int): год выхода альбома.

        genres (list[types.Genre], optional): жанры альбома.

        streams (int): количество прослушиваний альбома.

        saves (int): количество сохранений альбома пользователями или группами в свою музыку.

        uploadedAt (datetime): дата и время загрузки альбома (не релиза).

        updatedAt (datetime): дата и время обновления альбома (не релиза).

        exclusive (bool): флаг, указывающий, является ли альбом выпущенным эксклюзивно во ВКонтакте.

        photo (dict, optional): словарь с ссылками на различные размеры обложки альбома, отсортированные по возрастанию.

        original (types.Album, optional): оригинальный альбом, если это не альбом, а плейлист, сохранённый пользователем или группой в свою музыку.

        trackCount (int): количество треков в альбоме.

        tracks (list[types.Track]): треки альбома. Доступны при получении через client.getPlaylist(includeTracks=True).

        ownerId (int): идентификатор владельца альбома.

        albumId или playlistId (int): идентификатор альбома или плейлиста.

        id (str): полный идентификатор альбома в формате {ownerId}_{albumId or playlistId}.

        url (str): ссылка на альбом в формате https://vk.com/music/album/{id}.

        raw (dict): необработанные данные, полученные от ВКонтакте.
    """

    from typing import Union, List

    from vkmusix.aio import async_
    from vkmusix.types.track import Track

    def __init__(self, album: dict, playlist: bool = None, client: "Client" = None) -> None:
        import html

        from vkmusix.config import VK
        from vkmusix.utils import unixToDatetime

        from vkmusix.types.artist import Artist
        from vkmusix.types.genre import Genre

        super().__init__(client)

        title = album.get("title")
        self.title = html.unescape(title) if title else None

        subtitle = album.get("subtitle")
        self.subtitle = html.unescape(subtitle.replace("\n", " ")) if subtitle else None

        self.fullTitle = f"{self.title} ({self.subtitle})".replace("((", "(").replace("))", ")").replace("([", "(").replace("])", ")") if self.subtitle else self.title

        artists = self._client._finalizeResponse(
            album.get("main_artists"),
            Artist,
        )

        featuredArtists = self._client._finalizeResponse(
            album.get("featured_artists"),
            Artist,
        )

        self.artist = (
            (", ".join([
                artist.nickname
                for artist in artists[:-1]
            ]) + " & " + artists[-1].nickname)
            if len(artists) > 1 else
            artists[0].nickname +

            (
                (" feat. " + (
                    (", ".join([
                        artist.nickname
                        for artist in featuredArtists[:-1]
                    ]) + " & " + featuredArtists[-1].nickname)
                    if len(featuredArtists) > 1 else
                    featuredArtists[0].nickname))
                if featuredArtists else str()
            )
        ) if artists else None

        self.artists = artists
        self.featuredArtists = featuredArtists

        description = album.get("description")
        self.description = description if description else None

        self.releaseYear = album.get("year")

        self.genres = self._client._finalizeResponse(
            album.get("genres"),
            Genre,
        )

        self.streams = album.get("plays")
        self.saves = album.get("followers")

        self.uploadedAt = unixToDatetime(album.get("create_time"))
        self.updatedAt = unixToDatetime(album.get("update_time"))

        self.exclusive = album.get("exclusive")

        photo = album.get("photo")
        if not photo:
            photo = album.get("thumb")
        self.photo = {int(key.split("_")[1]): value[:value.rfind("&c_uniq_tag=")][:value.rfind("&type=")] for key, value in photo.items() if key.startswith("photo_")} if photo else None

        self.original = self._client._finalizeResponse(
            album.get("original"),
            Album,
        )

        tracks = album.get("tracks")
        self.trackCount = album.get("count") or (len(tracks) if tracks else None)
        self.tracks = tracks

        self.ownerId = album.get("owner_id")
        if not playlist:
            self.playlistId = None
            self.albumId = album.get("id") or album.get("album_id") or album.get("playlist_id")
            self.id = f"{self.ownerId}_{self.albumId}"
            self.url = f"{VK}music/album/{self.id}"

        else:
            self.albumId = None
            self.playlistId = album.get("id") or album.get("playlist_id") or album.get("album_id")
            self.id = f"{self.ownerId}_{self.playlistId}"
            self.url = f"{VK}music/playlist/{self.id}"

        self.raw = album


    @async_
    async def get(self, includeTracks: bool = False) -> Union["Album", None]:
        """
        Получает информацию об альбоме.

        `Пример использования`:

        album = album.get(
            includeTracks=True,
        )

        print(album)

        :param includeTracks: флаг, указывающий, небходимо ли также получить треки. (``bool``, `optional`)
        :return: `При успехе`: информация об альбоме (``types.Album``). `Если альбом не найден`: ``None``.
        """

        return await self._client.getPlaylist(
            self.albumId or self.playlistId,
            self.ownerId,
            includeTracks,
        )


    @async_
    async def getTracks(self) -> Union[List[Track], None]:
        """
        Получает треки альбома.

        `Пример использования`:

        tracks = album.getTracks()

        print(tracks)

        :return: `При успехе`: треки альбома (``list[types.Track]``). `Если альбом не найден`: ``None``.
        """

        return await self._client.getPlaylistTracks(
            self.albumId or self.playlistId,
            self.ownerId,
        )


    @async_
    async def add(self, groupId: int = None) -> Union["Album", None]:
        """
        Добавляет альбом в музыку пользователя или группы.

        `Пример использования`:

        album = album.add()

        print(album)

        :param groupId: идентификатор группы, в которую необходимо добавить альбом. (``int``, `optional`)
        :return: `При успехе`: информация о добавленном альбоме (``types.Album``). `Если альбом не найден`: ``None``.
        """

        return await self._client.addPlaylist(
            self.albumId or self.playlistId,
            self.ownerId,
            groupId,
        )


    @async_
    async def remove(self, groupId: int = None, validateIds: bool = True) -> bool:
        """
        Удаляет альбом из музыки пользователя или группы.

        `Пример использования`:

        result = album.remove()

        print(result)

        :param groupId: идентификатор группы, из которой необходимо удалить альбом. (``int``, `optional`)
        :param validateIds: флаг, указывающий, необходимо ли перепроверить альбом на наличие в музыке. По умолчанию ``True``. Установите на ``False``, если вы получили альбом через ``client.getPlaylists()`` или ``client.getAllPlaylists()``. (``bool``, `optional`)
        :return: `При успехе`: ``True``. `Если альбом не найден или не получилось его удалить`: ``False``.
        """

        return await self._client.removePlaylist(
            self.albumId or self.playlistId,
            self.ownerId,
            groupId,
            validateIds,
        )


    @async_
    async def copy(self, groupId: int = None, chatId: int = None, title: Union[str, None] = str(), description: Union[str, None] = str(), photo: Union[str, None] = str()) -> Union["Playlist", None]:
        """
        Копирует альбом в музыку пользователя или группы.

        `Пример использования`:

        playlist = playlist.copy()

        print(playlist)

        :param groupId: идентификатор группы, в которую необходимо скопировать альбом. (``int``, `optional`)
        :param chatId: идентификатор чата, к которому необходимо привязать скопированный альбом. (``int``, `optional`)
        :param title: название альбома. ``None`` для удаления. (``Union[str, None]``, `optional`)
        :param description: описание альбома. ``None`` для удаления. (``Union[str, None]``, `optional`)
        :param photo: ссылка на фото альбома. ``None`` для удаления. Не для удаления не работает. (``Union[str, None]``, `optional`)
        :return: `При успехе`: информация о скопированном альбоме (``types.Album``). `Если альбом не найден`: ``None``.
        """

        return await self._client.copyPlaylist(
            self.playlistId,
            self.ownerId,
            groupId,
            chatId,
            title,
            description,
            photo,
        )


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

from json import JSONEncoder

class Encoder(JSONEncoder):
    def default(self, o) -> any:
        from datetime import datetime

        if any((class_.__name__ == "Base" for class_ in o.__class__.__bases__)):
            return o._toDict()

        elif isinstance(o, datetime):
            return o.strftime("%d/%m/%Y %H:%M:%S")

        elif isinstance(o, type):
            return o.__name__

        return super().default(o)


class Base:
    def __init__(self, client: "Client") -> None:
        self._client = client


    def _toDict(self) -> dict:
        from types import FunctionType, MethodType

        result = dict()
        for key, value in self.__dict__.items():
            if any((value is None, all((key == "fullTitle", not getattr(self, "subtitle", None))), isinstance(value, (FunctionType, MethodType)), key in ["_client", "raw"])):
                continue

            result[key if not key.startswith("_") else key[1:]] = value

        return result


    def __repr__(self) -> str:
        from json import dumps

        return dumps(self._toDict(), indent=4, ensure_ascii=False, cls=Encoder)


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

class Remove:
    from typing import Union, List

    from vkmusix.aio import async_

    @async_
    async def remove(self, ownerIds: Union[List[int], int], trackIds: Union[List[int], int], playlistId: int = None, groupId: int = None, validateIds: bool = True) -> List[bool]:
        """
        Удаляет треки из музыки или плейлиста пользователя или группы.

        `Пример использования`:

        result = client.remove(
            ownerIds=-2001471901,
            trackIds=123471901,
        )

        print(result)

        :param ownerIds: идентификатор владельца треков. (``Union[list[int], int]``)
        :param trackIds: идентификаторы треков. (``Union[list[int], int]``)
        :param playlistId: идентификатор плейлиста, из которого необходимо удалить треки. Метод не работает для плейлистов, привязанных к чату (``int``, `optional`)
        :param groupId: идентификатор группы, из музыки или плейлиста которой необходимо удалить треки. (``int``, `optional`)
        :param validateIds: флаг, указывающий, необходимо ли перепроверить треки на наличие в музыке или плейлисте. По умолчанию ``True``. Установите на ``False``, если вы получили треки через ``client.getTracks()`` (при удалении из музыки) или ``client.getPlaylistTracks()`` (при удалении из плейлиста). (``bool``, `optional`)
        :return: Статусы удаления треков (``list[bool]``). `При успехе`: ``True``. `Если трек не удалось удалить`: ``False``.
        """

        if type(ownerIds) != type(trackIds):
            self._raiseError("ownerIdsAndTrackIdsTypeDifferent")

        elif isinstance(ownerIds, list) and isinstance(trackIds, list) and len(ownerIds) != len(trackIds):
            self._raiseError("ownerIdsAndTrackIdsLenDifferent")

        if not all((isinstance(ownerIds, list), isinstance(trackIds, list))):
            ownerIds = [ownerIds]
            trackIds = [trackIds]

        if not groupId:
            groupId = await self._getMyId()

        if validateIds and playlistId: # and playlistId временно
            if playlistId:
                existTracks = await self.getPlaylistTracks(playlistId, groupId)

            else:
                existTracks = NotImplemented

            if not existTracks and playlistId:
                return [False] * len(ownerIds)

            for index, (ownerId, trackId) in enumerate(zip(ownerIds, trackIds)):
                if not playlistId:
                    continue

                track = await self.get(ownerId, trackId)

                trackTitle = track.title
                trackArtists = track.artists
                if not trackArtists:
                    trackArtists = track.artist

                for existTrack in existTracks:
                    existTrackTitle = existTrack.title
                    existTrackArtists = existTrack.artists
                    if not existTrackArtists:
                        existTrackArtists = existTrack.artist

                    existTrackOwnerId = existTrack.ownerId
                    existTrackId = existTrack.trackId

                    if existTrackTitle == trackTitle and existTrackArtists == trackArtists:
                        ownerIds[index] = existTrackOwnerId
                        trackIds[index] = existTrackId
                        break

        from asyncio import Semaphore, gather

        deleteSemaphore = Semaphore(2)
        if playlistId:
            if groupId > 0:
                groupId = -groupId

            async def removeTrackFromPlaylist(id: str, semaphore: Semaphore) -> bool:
                async with semaphore:
                    response = await self._req(
                        "removeFromPlaylist",
                        {
                            "audio_ids": id,
                            "owner_id": groupId,
                            "playlist_id": playlistId,
                        },
                    )

                    return bool(response)

            tasks = [removeTrackFromPlaylist(f"{ownerId}_{trackId}", deleteSemaphore) for ownerId, trackId in zip(ownerIds, trackIds)]

        else:
            async def removeTrack(ownerId: int, trackId: int, semaphore: Semaphore) -> bool:
                async with semaphore:
                    response = await self._req(
                        "delete",
                        {
                            "owner_id": ownerId,
                            "audio_id": trackId,
                            "group_id": groupId,
                        },
                    )

                    return bool(response)

            tasks = [removeTrack(ownerId, trackId, deleteSemaphore) for ownerId, trackId in zip(ownerIds, trackIds)]

        results = await gather(*tasks)

        return list(results)


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

class GetArtist:
    from typing import Union

    from vkmusix.aio import async_
    from vkmusix.types import Artist

    @async_
    async def getArtist(self, artistId: int, includeAlbums: bool = False, includeTracks: bool = False) -> Union[Artist, None]:
        """
        Получает информацию об артисте.

        `Пример использования`:

        artist = client.getArtist(
            artistId=5696274288194638935,
            includeAlbums=True,
            includeTracks=True,
        )

        print(artist)

        :param artistId: идентификатор артиста. (``int``)
        :param includeAlbums: флаг, указывающий, небходимо ли также получить альбомы. (``bool``, `optional`)
        :param includeTracks: флаг, указывающий, небходимо ли также получить треки. (``bool``, `optional`)
        :return: `При успехе`: информация об артисте (``types.Artist``). `Если артист не найден`: ``None``.
        """

        from asyncio import gather

        from vkmusix.types import Artist

        tasks = [self._req(
            "getArtistById",
            {
                "artist_id": artistId,
            },
        )]

        if includeAlbums:
            tasks.append(self.getArtistAlbums(artistId))

        if includeTracks:
            tasks.append(self.getArtistTracks(artistId))

        responses = await gather(*tasks)

        artist = responses[0]

        if not artist.get("name"):
            return

        if includeAlbums:
            artist["albums"] = responses[1]

        if includeTracks:
            artist["tracks"] = responses[2 if includeAlbums else 1]

        return self._finalizeResponse(artist, Artist)

    get_artist = getArtist


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

class Download:
    from typing import Union

    from vkmusix.aio import async_
    from vkmusix.types import Track
    from vkmusix.enums import Extension

    @async_
    async def download(self, ownerId: int = None, trackId: int = None, filename: str = None, directory: str = None, extension: Extension = None, metadata: bool = False, track: Track = None) -> Union[str, None]:
        """
        Скачивает трек.

        `Пример использования`:

        from vkmusix.enums import Extension

        path = client.download(
            ownerId=-2001471901,
            trackId=123471901,
            extension=Extension.OPUS,
            metadata=True,
        )

        print(path)

        :param ownerId: идентификатор владельца трека. (``int``)
        :param trackId: идентификатор трека. (``int``)
        :param filename: имя файла с треком. По умолчанию ``{artist} — {fullTitle}``. Поддерживаемые переменные для динамического имени: ``artist``, ``title``, ``subtitle``, ``fullTitle``, ``album``. Пример динамического имени файла: ``{artist} - {title} ({album})``. (``str``, `optional`)
        :param directory: путь к директории, в которую загрузить трек. (``str``, `optional`)
        :param extension: расширение файла с треком. По умолчанию ``Extension.MP3``. (``enums.Extension``, `optional`)
        :param metadata: флаг, указывающий, необходимо ли добавить метаданные (артист, название, альбом, обложка) к файлу с треком. По умолчанию ``False``. Игнорируется, если параметр ``extension`` равен ``Extension.TS``. (``bool``, `optional`)
        :param track: трек. (``types.Track``, `optional`)
        :return: `При успехе`: полный путь к загруженному файлу (``str``). `Если трек не найден или недоступен для загрузки`: ``None``.
        """

        from asyncio import create_task, gather

        import re

        import os
        import aiofiles
        import aiofiles.os

        from Crypto.Cipher import AES
        from Crypto.Util.Padding import pad

        import av

        from vkmusix.types import Track
        from vkmusix.enums import Extension

        if not any((all((ownerId, trackId)), all((track, isinstance(track, Track))))):
            return

        if not track or not track.fileUrl:
            if track:
                ownerId, trackId = track.ownerId, track.trackId

            track = await self.get(ownerId, trackId)
            if not track.fileUrl:
                return

        async def downloadSegment(segmentUrlLocal: str, keyLocal: bytes, ivLocal: bytes) -> None:
            segmentData = await self._client.req(segmentUrlLocal, responseType="response")
            while segmentData.status_code in (301, 302):
                segmentData = await self._client.req(segmentData.headers.get("Location"), responseType="response")

            if segmentData.status_code != 200:
                return

            segmentData = segmentData.content

            if keyLocal:
                if len(segmentData) % AES.block_size != 0:
                    segmentData = pad(segmentData, AES.block_size)

                decryptor = AES.new(keyLocal, AES.MODE_CBC, ivLocal)
                decryptedData = decryptor.decrypt(segmentData)
                segmentData = decryptedData

            return segmentData

        if not directory:
            directory = os.getcwd()

        else:
            os.makedirs(directory, exist_ok=True)

        extension = extension.value if extension and isinstance(extension, Extension) else "mp3"

        if not filename:
            filename = f"{track.artist} — {track.fullTitle}"

        else:
            filename = filename if not filename.endswith(f".{extension}") else filename[:-(len(extension) + 1)]

            if "{subtitle}" in filename and not track.subtitle:
                filename = filename.replace("{subtitle}", str())

            if "{album}" in filename and not track.album:
                filename = filename.replace("{album}", str())

            filename = filename.format(
                artist=track.artist,
                title=track.title,
                subtitle=track.subtitle,
                fullTitle=track.fullTitle,
                album=track.album.title if track.album else None,
            )

        filename = re.sub(r'[<>:"/\\|?*]', str(), filename)
        filename = os.path.join(directory, filename)

        key = None
        iv = bytes.fromhex("00000000000000000000000000000000")
        tasks = list()

        m3u8Content = await self._client.req(track.fileUrl, responseType="response")
        while m3u8Content.status_code in (301, 302):
            m3u8Content = await self._client.req(m3u8Content.headers.get("Location"), responseType="response")

        if m3u8Content.status_code != 200:
            return

        for line in m3u8Content.text.splitlines():
            if line.startswith("#EXT-X-KEY"):
                method = line.split("METHOD=")[1].split(",")[0]

                if method == "AES-128":
                    if not key:
                        keyUri = line.split('URI="')[1].split('"')[0]
                        response = await self._client.req(keyUri, responseType="response")
                        contentType = response.headers.get("content-type")
                        key = response.content if contentType and contentType == "application/octet-stream" else response.text.encode()

                    keyLocal = key

                elif method == "NONE":
                    keyLocal = None

            elif line.endswith((".ts", ".ts?siren=1")):
                segmentUrl = os.path.join(os.path.dirname(track.fileUrl), line)
                task = create_task(downloadSegment(str(segmentUrl), keyLocal, iv))
                tasks.append(task)

        segments = await gather(*tasks)

        async with aiofiles.open(f"{filename}.ts", "wb") as outfile:
            buffer = bytearray()
            for segment in segments:
                if not segment:
                    continue

                buffer.extend(segment)

                if len(buffer) > 10 * 1024 * 1024:
                    await outfile.write(buffer)
                    buffer.clear()

            if buffer:
                await outfile.write(buffer)

        if extension != "ts":
            try:
                inputContainer = av.open(f"{filename}.ts")
                outputContainer = av.open(f"{filename}.{extension}", mode="w", format=extension)

                inputStream = inputContainer.streams.audio[0]
                outputStream = outputContainer.add_stream(extension, rate=inputStream.rate)

                if extension == "opus":
                    outputStream.codec_context.options = {
                        "strict": "experimental",
                    }

                for packet in inputContainer.demux(inputStream):
                    if packet.stream == inputStream and packet.stream_index == inputStream.index:
                        if extension == "opus":
                            if packet.size > 0:
                                try:
                                    for frame in packet.decode():
                                        newPacket = outputStream.encode(frame)

                                        if newPacket:
                                            outputContainer.mux(newPacket)

                                except av.error.InvalidDataError:
                                    pass
                        else:
                            outputContainer.mux(packet)

                inputContainer.close()
                outputContainer.close()

            except (av.InvalidDataError, av.ValueError):
                return

            finally:
                await aiofiles.os.remove(f"{filename}.ts")

            if metadata:
                album = track.album
                photo = album.photo if album else None

                coverData = await self._client.req(
                    photo.get(1200) or photo.get(600) or photo.get(300) or photo.get(270),
                    responseType="file",
                ) if photo else None

                if extension == "mp3":
                    from mutagen.id3 import ID3, APIC, TIT2, TPE1, TALB
                    from mutagen.mp3 import MP3

                    audio = MP3(f"{filename}.{extension}", ID3=ID3)
                    audio.update(
                        {
                            **{
                                "TIT2": TIT2(encoding=3, text=[track.fullTitle]),
                                "TPE1": TPE1(encoding=3, text=[track.artist]),
                            },
                            **({
                                "TALB": TALB(encoding=3, text=[album.title]),
                            } if album else dict()),
                        },
                    )

                    if coverData:
                        audio.tags.add(
                            APIC(
                                encoding=3,
                                mime="image/jpeg",
                                type=3,
                                desc=u"Cover",
                                data=coverData,
                            )
                        )

                    audio.save()

                elif extension == "opus":
                    from mutagen.oggopus import OggOpus

                    audio = OggOpus(f"{filename}.{extension}")
                    audio.update(
                        {
                            "title": track.fullTitle,
                            "artist": track.artist,
                            **({
                                "album": album.title,
                            } if album else dict()),
                        },
                    )

                    if coverData:
                        import base64
                        from mutagen.flac import Picture

                        picture = Picture()

                        picture.data = coverData

                        picture.type = 3
                        picture.mime = "image/jpeg"

                        resolution = next((x for x in [1200, 600, 300, 270] if photo.get(x)), 1200)
                        picture.width = resolution
                        picture.height = resolution

                        picture.depth = 24

                        encodedPicture = base64.b64encode(picture.write()).decode("ascii")
                        audio["metadata_block_picture"] = [encodedPicture]

                    audio.save()

        return f"{filename}.{extension}"


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

class SearchTracks:
    from typing import Union, List

    from vkmusix.aio import async_
    from vkmusix.types import Track

    @async_
    async def searchTracks(self, query: str, limit: int = None, offset: int = None) -> Union[List[Track], None]:
        """
        Ищет треки.

        `Пример использования`:

        tracks = client.searchTracks(
            query="Маленький ярче — LARI LARI",
            limit=10,
        )

        print(tracks)

        :param query: поисковой запрос. (``str``)
        :param limit: лимит треков. (``int``, `optional`)
        :param offset: сколько треков пропустить. (``int``, `optional`)
        :return: `При успехе`: найденные треки (``list[types.Track]``). `Если треки не найдены`: ``None``.
        """

        from vkmusix.types import Track

        return await self._search(
            "search",
            (query, limit, offset),
            Track,
        )

    search_tracks = searchTracks


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

class CreatePlaylist:
    from typing import Union

    from vkmusix.aio import async_
    from vkmusix.types import Playlist

    @async_
    async def createPlaylist(self, title: str, description: str = None, photo: str = None, groupId: int = None, chatId: int = None) -> Union[Playlist, None]:
        """
        Создаёт плейлист в музыке пользователя или группы.

        `Пример использования`:

        playlist = client.createPlaylist(
            title="Лучшая музыка в машину!!!",
        )

        print(playlist)

        :param title: название плейлиста. (``str``)
        :param description: описание плейлиста. (``str``, `optional`)
        :param photo: ссылка на фото плейлиста. Не работает. (``str``, `optional`)
        :param groupId: идентификатор группы, в которой необходимо создать плейлист. (``int``, `optional`)
        :param chatId: идентификатор чата, к которому необходимо привязать плейлист. (``int``, `optional`)
        :return: `При успехе`: информация о созданном плейлисте (``types.Playlist``). `Если плейлист не удалось создать`: ``None``.
        """

        from vkmusix.types import Playlist

        if not groupId:
            groupId = await self._getMyId()

        playlist = await self._req(
            "createPlaylist" if not chatId else "createChatPlaylist",
            {
                "title": title,
                "description": description,
                "owner_id": groupId
            }
        )

        playlistId = playlist.get("id")
        if chatId:
            groupId = playlist.get("owner_id")

        """if photo:
            await self._editPlaylistPhoto(playlistId, photo, groupId)"""

        return Playlist(
            {
                "owner_id": groupId,
                "playlist_id": playlistId,
            },
            True,
            self,
        )

    create_playlist = createPlaylist


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

class Edit:
    from vkmusix.aio import async_

    @async_
    async def edit(self, ownerId: int, trackId: int, title: str = None, artist: str = None, lyrics: str = None, genreId: int = None, removeFromSearchResults: bool = None) -> bool:
        """
        Изменяет информацию о треке.

        `Пример использования`:

        result = client.edit(
            ownerId=-2001471901,
            trackId=123471901,
            filename="Маленький ярче — LARILARI",
            title="LARILARI",
            artist="Маленький ярче",
            genreId=21,
            removeFromSearchResults=True,
        )

        print(result)

        :param ownerId: идентификатор владельца трека. (``int``)
        :param trackId: идентификатор трека. (``int``)
        :param title: название трека. По умолчанию берётся из метаданных файла. (``str``, `optional`)
        :param artist: артисты трека. По умолчанию берётся из метаданных файла. (``str``, `optional`)
        :param lyrics: текст трека. (``str``, `optional`)
        :param genreId: идентификатор жанра трека. (``int``, `optional`)
        :param removeFromSearchResults: флаг, указывающий, необходимо ли исключить трек из поиска. По умолчанию ``False``. (``bool``, `optional`)
        :return: `При успехе`: ``True``. `Если информацию о треке не удалось изменить`: ``False``.
        """

        if not any((title, artist, lyrics is not None, genreId is not None, removeFromSearchResults is not None)):
            return False

        response = await self._req(
            "edit",
            {
                "owner_id": ownerId,
                "audio_id": trackId,
                **({"title": title} if title else dict()),
                **({"artist": artist} if artist else dict()),
                **({"lyrics": lyrics} if lyrics is not None else dict()),
                **({"genre_id": genreId} if genreId is not None else dict()),
                **({"no_search": removeFromSearchResults} if removeFromSearchResults is not None else dict()),
            },
        )

        return bool(response)


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

class Add:
    from typing import Union, List

    from vkmusix.aio import async_

    @async_
    async def add(self, ownerIds: Union[List[int], int], trackIds: Union[List[int], int], playlistId: int = None, groupId: int = None) -> List[bool]:
        """
        Добавляет треки в музыку или плейлист пользователя или группы.

        `Пример использования`:

        result = client.add(
            ownerIds=-2001471901,
            trackIds=123471901,
        )

        print(result)

        :param ownerIds: идентификаторы владельцев треков. (``Union[list[int], int]``)
        :param trackIds: идентификаторы треков. (``Union[list[int], int]``)
        :param playlistId: идентификатор плейлиста, в который необходимо добавить треки. (``int``, `optional`)
        :param groupId: идентификатор группы, в музыку или плейлист которой необходимо добавить треки. (``int``, `optional`)
        :return: Статусы добавления треков (``list[bool]``). `При успехе`: ``True``. `Если трек не удалось добавить`: ``False``.
        """

        if type(ownerIds) != type(trackIds):
            self._raiseError("ownerIdsAndTrackIdsTypeDifferent")

        elif isinstance(ownerIds, list) and isinstance(trackIds, list) and len(ownerIds) != len(trackIds):
            self._raiseError("ownerIdsAndTrackIdsLenDifferent")

        if not all((isinstance(ownerIds, list), isinstance(trackIds, list))):
            ownerIds = [ownerIds]
            trackIds = [trackIds]

        if not groupId:
            groupId = await self._getMyId()

        method = "add"
        if playlistId:
            method += "ToPlaylist"

        results = list()
        for ownerId, trackId in zip(ownerIds, trackIds):
            response = await self._req(
                method,
                {
                    **(
                        {
                            "owner_id": ownerId,
                            "audio_id": trackId,
                            "group_id": groupId,
                        }
                        if not playlistId else
                        {
                            "owner_id": groupId,
                            "audio_ids": f"{ownerId}_{trackId}",
                        }
                    ),
                    "playlist_id": playlistId,
                },
            )

            results.append(bool(response))

        return results


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

class RemovePlaylist:
    from vkmusix.aio import async_

    @async_
    async def removePlaylist(self, playlistId: int, ownerId: int = None, groupId: int = None, validateIds: bool = True) -> bool:
        """
        Удаляет плейлист или альбом из музыки пользователя или группы.

        `Пример использования`:

        result = client.removePlaylist(
            playlistId=19201020,
            ownerId=-2000201020,
        )

        print(result)

        :param playlistId: идентификатор плейлиста или альбома. (``int``)
        :param ownerId: идентификатор владельца плейлиста или альбома (пользователь или группа). (``int``, `optional`)
        :param groupId: идентификатор группы, из которой необходимо удалить плейлист или альбом. (``int``, `optional`)
        :param validateIds: флаг, указывающий, необходимо ли перепроверить плейлист или альбом на наличие в музыке. По умолчанию ``True``. Установите на ``False``, если вы получили плейлист или альбом через ``client.getPlaylists()`` или ``client.getAllPlaylists()``. (``bool``, `optional`)
        :return: `При успехе`: ``True``. `Если плейлист или альбом не найден или не получилось его удалить`: ``False``.
        """

        from vkmusix.errors import NotFound

        if not ownerId:
            ownerId = await self._getMyId()

        if not groupId:
            groupId = await self._getMyId()

        if validateIds:
            from vkmusix.enums import PlaylistType

            existPlaylists = await self.getAllPlaylists(
                groupId,
                [
                    PlaylistType.Foreign,
                    PlaylistType.Album,
                ],
            )

            if not existPlaylists:
                return False

            playlist = await self.getPlaylist(
                playlistId,
                ownerId,
            )

            for existPlaylist in existPlaylists:
                original = existPlaylist.original

                if not original:
                    continue

                if original.id == playlist.id:
                    playlistId = existPlaylist.playlistId or existPlaylist.albumId
                    break

        try:
            result = await self._req(
                "deletePlaylist",
                {
                    "owner_id": groupId,
                    "playlist_id": playlistId,
                },
            )

        except NotFound:
            return False

        return bool(result)

    remove_playlist = removePlaylist


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

import asyncio
import httpx
import base64

from typing import Union, List, Type

from vkmusix.aio import async_
from vkmusix.config import VKAPI, VKAPIVersion, RuCaptchaAPI
from vkmusix.errors import *

from vkmusix.utils import checkFile
from vkmusix.webClient import Client as WebClient

from vkmusix.methods import Methods
from vkmusix.enums import Language

class Client(Methods):
    """
    Класс, который вы будете использовать для каждого взаимодействия с VK Music.

    Параметры:
        token (str, optional): Токен доступа к ВКонтакте API с правами на аудио.\n
        RuCaptchaKey (str, optional): Ключ доступа к RuCaptcha API для автоматического решения капч через этот сервис. Если не указан, капча потребует ручного решения.\n
        language (enums.Language, optional): Язык ошибок (например, Language.Russian для русского, Language.English для английского). Если не указан, используются все языки.\n
        proxies (dict, optional): Прокси, которые будут использоваться при запросах. Формат: {"http": "IP:port"} или {"socks5": "login:password@IP:port"}.\n

    Создания экземпляра:
        from vkmusix import Client
        from vkmusix.enums import Language

        client = Client(
            token="...",
            RuCaptchaKey="...",
            language=Language.Russian or Language.English,
            proxies={
                "http": "IP:port",
                "socks5": "login:password@IP:port",
            },
        ) # Все параметры необязательны

        # Или с использованием with

        with Client(
            token="...",
            RuCaptchaKey="...",
            language=Language.Russian or Language.English,
            proxies={
                "http": "IP:port",
                "socks5": "login:password@IP:port",
            },
        ) as client:
            ...
    """


    def __init__(self, token: str = None, RuCaptchaKey: str = None, language: Language = None, proxies: dict = None) -> None:
        import sys
        if sys.version_info < (3, 6):
            ruWarning = "Внимание: Работоспособость этой библиотеки гарантируется только при Python 3.6 или выше. Вы используете версию {}."
            enWarning = "Attention: The functionality of this library is guaranteed only for Python 3.6 or higher. You are using version {}."

            from warnings import warn
            warn(
                (ruWarning if self._language == Language.Russian else (enWarning if self._language == Language.English else "🇷🇺: " + ruWarning + " 🇬🇧: " + enWarning)).format(str(sys.version_info.major) + "." + str(sys.version_info.minor)),
                UserWarning,
            )

        if not token:
            token = input("Получите токен ВКонтакте с правами на аудиозаписи и доступ в любое время на сайте `https://vkhost.github.io/` (приложение VK Admin) и отправьте его: ")

        self._RuCaptchaKey = RuCaptchaKey
        self._language = language if language and isinstance(language, Language) else None

        if not proxies:
            self._proxies = None

        else:
            if not isinstance(proxies, dict):
                self._raiseError("proxyShouldBeDict")

            newProxies = dict()

            for scheme in proxies.keys():
                if scheme.lower() not in ("http", "https", "socks4", "socks5"):
                    self._raiseError("invalidProxyDict")

                proxyURL = proxies.get(scheme)
                scheme = scheme.lower() + ("://" if not scheme.endswith("://") else str())
                newProxies[scheme] = (scheme if "://" not in proxyURL else str()) + proxyURL

            self._proxies = newProxies

        self._session = httpx.AsyncClient(proxies=self._proxies)
        self._client = WebClient(self._session)

        self._params = {
            "access_token": token,
            "v": VKAPIVersion,
        }
        self._closed = False
        self._me = None

        try:
            asyncio.get_running_loop()

        except RuntimeError:
            from .version import __version__
            from packaging import version

            latestVersion = self._client.req("https://pypi.org/pypi/vkmusix/json").get("info").get("version")

            if version.parse(latestVersion) > version.parse(__version__):
                ruWarning = f"Внимание: Доступна новая версия библиотеки {latestVersion} (https://pypi.org/project/vkmusix). Вы используете версию {__version__}."
                enWarning = f"Attention: A new version of the library {latestVersion} (https://pypi.org/project/vkmusix) is available. You are using version {__version__}."

                from warnings import warn
                warn(
                    ruWarning if self._language == Language.Russian else (enWarning if self._language == Language.English else "🇷🇺: " + ruWarning + " 🇬🇧: " + enWarning),
                    UserWarning,
                )


    @async_
    async def checkUpdates(self) -> None:
        if self._closed:
            self._raiseError("sessionClosed")

        from .version import __version__
        from packaging import version

        latestVersion = (await self._client.req("https://pypi.org/pypi/vkmusix/json")).get("info").get("version")

        if version.parse(latestVersion) > version.parse(__version__):
            ruWarning = f"Внимание: Доступна новая версия библиотеки {latestVersion} (https://pypi.org/project/vkmusix). Вы используете версию {__version__}."
            enWarning = f"Attention: A new version of the library {latestVersion} (https://pypi.org/project/vkmusix) is available. You are using version {__version__}."

            from warnings import warn
            warn(
                ruWarning if self._language == Language.Russian else (enWarning if self._language == Language.English else "🇷🇺: " + ruWarning + " 🇬🇧: " + enWarning),
                UserWarning,
            )


    def __enter__(self) -> "Client":
        return self


    async def __aenter__(self) -> "Client":
        return self


    def __exit__(self, exc_type, exc_value, traceback) -> None:
        try:
            self.close()

        except SessionAlreadyClosed:
            pass


    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        try:
            await self.close()

        except SessionAlreadyClosed:
            pass


    @async_
    async def close(self) -> None:
        """
        Закрывает текующую сессию. Для отправки новых запросов потребуется создать новый объект класса `Client` или использовать метод `client.reconnect()`.
        """

        if self._closed:
            self._raiseError("sessionAlreadyClosed")
            return

        self._closed = True
        await self._session.aclose()


    @async_
    async def reconnect(self) -> None:
        """
        Пересоздаёт закрытую сессию.
        """

        if not self._closed:
            self._raiseError("sessionAlreadyOpened")
            return

        self._closed = False
        self._session = httpx.AsyncClient(proxies=self._proxies)
        self._client = WebClient(self._session)


    @async_
    async def _getMyId(self) -> int:
        if not self._me:
            self._me = await self.getMe()

        return self._me.get("id")


    @async_
    async def _req(self, method: str, params: dict = None, json: dict = None, data: any = None, cookies: dict = None, headers: dict = None, files: dict = None, version: float = None, httpMethod: str = None) -> Union[dict, None]:
        if self._closed:
            self._raiseError("sessionClosed")

        if not params:
            params = dict()

        else:
            params = {
                k: v
                for k, v in params.items()
                if v is not None
            }

            limit = params.get("count")

            if limit:
                if not isinstance(limit, int):
                    params["count"] = None

                elif limit < 0:
                    params["count"] = 300

        if "." not in method:
            method = f"audio.{method}"

        url = f"{VKAPI}{method}"
        fullParams = {**params, **self._params}

        if version:
            fullParams["v"] = version

        req = await self._client.req(
            url,
            fullParams,
            json,
            data,
            cookies,
            headers,
            files,
            method=httpMethod.upper() if httpMethod and isinstance(httpMethod, str) else "GET",
        )

        while isinstance(req, dict) and req.get("error"):
            error = req.get("error")
            errorCode = error.get("error_code")

            if errorCode == 3:
                self._raiseError("invalidMethod")

            elif errorCode == 5:
                self._raiseError("VKInvalidToken")

            elif errorCode in [6, 9]:
                self._raiseError("tooHighRequestSendingRate")

            elif errorCode == 10 and method == "audio.createChatPlaylist":
                self._raiseError("chatNotFound")

            elif errorCode == 14:
                captchaUrl = error.get("captcha_img")
                if self._RuCaptchaKey:
                    solve = await self._solveCaptcha(captchaUrl)

                else:
                    solve = input(captchaUrl + "\nВведите решение капчи: ")

                fullParams.update(
                    {
                        "captcha_sid": error.get("captcha_sid"),
                        "captcha_key": solve,
                    }
                )

                req = await self._client.req(url, fullParams)

            elif errorCode in [15, 201, 203]:
                if ": can not restore too late" in error.get("error_msg"):
                    self._raiseError("trackRestorationTimeEnded")

                else:
                    self._raiseError("accessDenied")

            elif errorCode == 18:
                self._raiseError("userWasDeletedOrBanned")

            elif errorCode == 100:
                if "One of the parameters" in error.get("error_msg"):
                    return error

                return

            elif errorCode == 104:
                if method == "audio.getLyrics":
                    return

                self._raiseError("notFound")

            else:
                return error

        if isinstance(req, list) and len(req) == 1:
            req = req[0]

        return req


    @async_
    async def _solveCaptcha(self, captchaUrl: str) -> str:
        imageBytes = await self._client.req(captchaUrl, responseType="file")
        captchaImageInBase64 = base64.b64encode(imageBytes).decode("utf-8")

        RuCaptchaParams = {
            "clientKey": self._RuCaptchaKey,
            "task": {
                "type": "ImageToTextTask",
                "body": captchaImageInBase64,
            },
            "languagePool": "rn",
        }

        taskId = (await self._client.req(
            f"{RuCaptchaAPI}createTask",
            json=RuCaptchaParams,
            method="POST",
        )).get("taskId")

        while True:
            await asyncio.sleep(5)
            taskResult = await self._client.req(
                f"{RuCaptchaAPI}getTaskResult",
                json={
                    "clientKey": self._RuCaptchaKey,
                    "taskId": taskId,
                },
                method="POST",
            )
            errorId = taskResult.get("errorId")

            if errorId == 0 and taskResult.get("status") == "ready":
                return taskResult.get("solution").get("text")

            elif errorId == 1:
                self._raiseError("RuCaptchaInvalidKey")

            elif errorId == 10:
                self._raiseError("RuCaptchaZeroBalance")

            elif errorId == 12:
                taskId = (await self._client.req(
                    f"{RuCaptchaAPI}createTask",
                    json=RuCaptchaParams,
                    method="POST",
                )).get("taskId")

            elif errorId == 21:
                self._raiseError("RuCaptchaBannedIP")

            elif errorId == 55:
                self._raiseError("RuCaptchaBannedAccount")


    def _raiseError(self, errorType: Union[str, None]) -> None:
        if not errorType:
            return

        errorsDict = {
            "unknown": Unknown,

            "sessionClosed": SessionClosed,
            "sessionAlreadyClosed": SessionAlreadyClosed,
            "sessionAlreadyOpened": SessionAlreadyOpened,

            "VKInvalidToken": VKInvalidToken,
            "VKCookieFileNotFound": VKCookieFileNotFound,
            "VKInvalidCookie": VKInvalidCookie,
            "VKUnsuccessfulLoginAttempt": VKUnsuccessfulLoginAttempt,

            "RuCaptchaInvalidKey": RuCaptchaInvalidKey,
            "RuCaptchaZeroBalance": RuCaptchaZeroBalance,
            "RuCaptchaBannedIP": RuCaptchaBannedIP,
            "RuCaptchaBannedAccount": RuCaptchaBannedAccount,

            "invalidMethod": InvalidMethod,
            "accessDenied": AccessDenied,
            "accessDeniedWithoutCookie": AccessDeniedWithoutCookie,

            "userWasDeletedOrBanned": UserWasDeletedOrBanned,
            "trackRestorationTimeEnded": TrackRestorationTimeEnded,

            "notFound": NotFound,
            "chatNotFound": ChatNotFound,

            "noneQuery": NoneQuery,

            "ownerIdsAndTrackIdsTypeDifferent": OwnerIdsAndTrackIdsTypeDifferent,
            "ownerIdsAndTrackIdsLenDifferent": OwnerIdsAndTrackIdsLenDifferent,

            "trackReorderNeedsBeforeOrAfterArgument": TrackReorderNeedsBeforeOrAfterArgument,
            "trackReorderNeedsOnlyBeforeOrAfterNotBoth": TrackReorderNeedsOnlyBeforeOrAfterNotBoth,

            "MP3FileNotFound": MP3FileNotFound,
            "MP3FileTooBig": MP3FileTooBig,

            "tooHighRequestSendingRate": TooHighRequestSendingRate,

            "invalidProxyType": InvalidProxyType,
            "invalidProxyDict": InvalidProxyDict,
        }

        if errorType not in errorsDict:
            errorType = "unknown"

        error = errorsDict.get(errorType)()

        if self._language:
            if self._language == Language.Russian:
                languageAttr = "ru"

            elif self._language == Language.English:
                languageAttr = "en"

            for attr in ["ru", "en"]:
                if attr != languageAttr:
                    delattr(error, attr)

        raise error


    def _finalizeResponse(self, response: Union[List[dict], dict], objectType: Type[any]) -> Union[List[any], None]:
        if not (response or response is False):
            return

        wasList = True
        if not isinstance(response, list):
            wasList = False
            response = [response]

        for index, obj in enumerate(response):
            from vkmusix.types import Album, Playlist

            if objectType in [Album, Playlist]:
                type_ = obj.get("type")

                if type_ in [0, 5]:
                    obj = Playlist(obj, False if obj.get("original") else True, client=self)

                elif type_ == 1:
                    obj = Album(obj, True, client=self)

                else:
                    obj = objectType(obj, client=self)

            else:
                obj = objectType(obj, client=self)

            response[index] = obj

        return response if wasList else response[0]


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

import re
import asyncio

import httpx

retries = 5
timeout = 20
sleepTime = .25

def addHTTPsToUrl(url: str) -> str:
    if not ("https://" in url or "http://" in url):
        url = "https://" + url

    return url

class Client:
    from typing import Union

    from vkmusix.aio import async_

    def __init__(self, client: httpx.AsyncClient) -> None:
        self.client = client


    @async_
    async def request(self, url: str, params: dict = None, json: dict = None, data: Union[str, dict] = None, cookies: dict = None, headers: dict = None, files: dict = None, responseType: str = "json", method: str = "GET") -> any:
        responseType = responseType.lower()
        retriesLocal = retries

        url = addHTTPsToUrl(url)

        if cookies and isinstance(cookies, list):
            cookies_ = dict()
            for index, cookie in enumerate(cookies):
                cookies_[cookie.get("name")] = cookie.get("value")
            cookies = cookies_

        while retriesLocal > 0:
            try:
                response = await self.client.request(
                    method,
                    url,
                    params=params,
                    json=json,
                    data=data,
                    cookies=cookies,
                    headers=headers,
                    files=files,
                    timeout=httpx.Timeout(timeout),
                    follow_redirects=False,
                )

                if responseType == "json":
                    responseJson = response.json()
                    if "response" in responseJson:
                        responseJson = responseJson.get("response")

                    return responseJson

                elif responseType == "code":
                    responseText = response.text
                    return re.sub(r"<br/>", "\n", responseText)

                elif responseType == "file":
                    return response.content

                elif responseType == "response":
                    return response

            except (httpx.TimeoutException, httpx.ConnectError, httpx.RequestError, httpx.ReadError):
                retriesLocal -= 1
                await asyncio.sleep(sleepTime)

            except asyncio.TimeoutError:
                retriesLocal -= 1
                await asyncio.sleep(sleepTime)

    req = request